# -*- coding: utf-8 -*-
"""
Regularized Unbalanced OT solvers
"""

# Author: Hicham Janati <hicham.janati@inria.fr>
#         Laetitia Chapel <laetitia.chapel@univ-ubs.fr>
#         Quang Huy Tran <quang-huy.tran@univ-ubs.fr>
#
# License: MIT License

from __future__ import division
import warnings

from ..backend import get_backend
from ..utils import list_to_array, get_parameter_pair


def sinkhorn_unbalanced(
    a,
    b,
    M,
    reg,
    reg_m,
    method="sinkhorn",
    reg_type="kl",
    c=None,
    warmstart=None,
    numItermax=1000,
    stopThr=1e-6,
    verbose=False,
    log=False,
    **kwargs,
):
    r"""
    Solve the unbalanced entropic regularization optimal transport problem
    and return the OT plan

    The function solves the following optimization problem:

    .. math::
        W = \arg \min_\gamma \ \langle \gamma, \mathbf{M} \rangle_F +
        \mathrm{reg} \cdot \mathrm{KL}(\gamma, \mathbf{c}) +
        \mathrm{reg_{m1}} \cdot \mathrm{KL}(\gamma \mathbf{1}, \mathbf{a}) +
        \mathrm{reg_{m2}} \cdot \mathrm{KL}(\gamma^T \mathbf{1}, \mathbf{b})

        s.t.
             \gamma \geq 0

    where :

    - :math:`\mathbf{M}` is the (`dim_a`, `dim_b`) metric cost matrix
    - :math:`\mathbf{a}` and :math:`\mathbf{b}` are source and target unbalanced distributions
    - :math:`\mathbf{c}` is a reference distribution for the regularization
    - KL is the Kullback-Leibler divergence

    The algorithm used for solving the problem is the generalized
    Sinkhorn-Knopp matrix scaling algorithm as proposed in :ref:`[10, 25]
    <references-sinkhorn-unbalanced>`

    .. warning::
        Starting from version 0.9.5, the default value has been changed to `reg_type='kl'` instead of `reg_type='entropy'`. This makes the function more consistent with the literature
        and the other solvers. If you want to use the entropy regularization, please set `reg_type='entropy'` explicitly.


    Parameters
    ----------
    a : array-like, shape (dim_a,)
        Unnormalized histogram of dimension `dim_a`
        If `a` is an empty list or array ([]),
        then `a` is set to uniform distribution.
    b : array-like, shape (dim_b,)
        One or multiple unnormalized histograms of dimension `dim_b`.
        If `b` is an empty list or array ([]),
        then `b` is set to uniform distribution.
        If many, compute all the OT costs :math:`(\mathbf{a}, \mathbf{b}_i)_i`
    M : array-like, shape (dim_a, dim_b)
        loss matrix
    reg : float
        Entropy regularization term > 0
    reg_m: float or indexable object of length 1 or 2
        Marginal relaxation term.
        If :math:`\mathrm{reg_{m}}` is a scalar or an indexable object of length 1,
        then the same :math:`\mathrm{reg_{m}}` is applied to both marginal relaxations.
        The entropic balanced OT can be recovered using :math:`\mathrm{reg_{m}}=float("inf")`.
        For semi-relaxed case, use either
        :math:`\mathrm{reg_{m}}=(float("inf"), scalar)` or
        :math:`\mathrm{reg_{m}}=(scalar, float("inf"))`.
        If :math:`\mathrm{reg_{m}}` is an array,
        it must have the same backend as input arrays `(a, b, M)`.
    method : str
        method used for the solver either 'sinkhorn', 'sinkhorn_stabilized', 'sinkhorn_translation_invariant' or
        'sinkhorn_reg_scaling', see those function for specific parameters
    reg_type : string, optional
        Regularizer term. Can take two values:

        - Negative entropy: 'entropy':
          :math:`\Omega(\gamma) = \sum_{i,j} \gamma_{i,j} \log(\gamma_{i,j}) - \sum_{i,j} \gamma_{i,j}`.
          This is equivalent (up to a constant) to :math:`\Omega(\gamma) = \text{KL}(\gamma, 1_{dim_a} 1_{dim_b}^T)`.
        - Kullback-Leibler divergence (default): 'kl':
          :math:`\Omega(\gamma) = \text{KL}(\gamma, \mathbf{a} \mathbf{b}^T)`.
    c : array-like, shape (dim_a, dim_b), optional (default=None)
        Reference measure for the regularization.
        If None, then use :math:`\mathbf{c} = \mathbf{a} \mathbf{b}^T`.
        If :math:`\texttt{reg_type}=`'entropy', then :math:`\mathbf{c} = 1_{dim_a} 1_{dim_b}^T`.
    warmstart: tuple of arrays, shape (dim_a, dim_b), optional
        Initialization of dual potentials. If provided, the dual potentials should be given
        (that is the logarithm of the `u`, `v` sinkhorn scaling vectors).
    numItermax : int, optional
        Max number of iterations
    stopThr : float, optional
        Stop threshold on error (>0)
    verbose : bool, optional
        Print information along iterations
    log : bool, optional
        record `log` if `True`


    Returns
    -------
    if n_hists == 1:
        - gamma : array-like, shape(dim_a, dim_b)
            Optimal transportation matrix for the given parameters
        - log : dict
            log dictionary returned only if `log` is `True`
    else:
        - ot_distance : array-like, shape (n_hists,)
            the OT distance between :math:`\mathbf{a}` and each of the histograms :math:`\mathbf{b}_i`
        - log : dict
            log dictionary returned only if `log` is `True`

    Examples
    --------

    >>> import ot
    >>> import numpy as np
    >>> a=[.5, .5]
    >>> b=[.5, .5]
    >>> M=[[0., 1.], [1., 0.]]
    >>> np.round(ot.sinkhorn_unbalanced(a, b, M, 1, 1), 7)
    array([[0.3220536, 0.1184769],
           [0.1184769, 0.3220536]])


    .. _references-sinkhorn-unbalanced:
    References
    ----------
    .. [2] M. Cuturi, Sinkhorn Distances : Lightspeed Computation of Optimal
        Transport, Advances in Neural Information Processing Systems
        (NIPS) 26, 2013

    .. [9] Schmitzer, B. (2016). Stabilized Sparse Scaling Algorithms for
        Entropy Regularized Transport Problems. arXiv preprint arXiv:1610.06519.

    .. [10] Chizat, L., Peyré, G., Schmitzer, B., & Vialard, F. X. (2016).
        Scaling algorithms for unbalanced transport problems. arXiv preprint
        arXiv:1607.05816.

    .. [25] Frogner C., Zhang C., Mobahi H., Araya-Polo M., Poggio T. :
        Learning with a Wasserstein Loss,  Advances in Neural Information
        Processing Systems (NIPS) 2015

    .. [73] Séjourné, T., Vialard, F. X., & Peyré, G. (2022).
       Faster unbalanced optimal transport: Translation invariant sinkhorn and 1-d frank-wolfe.
       In International Conference on Artificial Intelligence and Statistics (pp. 4995-5021). PMLR.


    See Also
    --------
    ot.unbalanced.sinkhorn_knopp_unbalanced: Unbalanced Classic Sinkhorn :ref:`[10] <references-sinkhorn-unbalanced>`
    ot.unbalanced.sinkhorn_stabilized_unbalanced:
        Unbalanced Stabilized sinkhorn :ref:`[9, 10] <references-sinkhorn-unbalanced>`
    ot.unbalanced.sinkhorn_reg_scaling_unbalanced:
        Unbalanced Sinkhorn with epsilon scaling :ref:`[9, 10] <references-sinkhorn-unbalanced>`
    ot.unbalanced.sinkhorn_unbalanced_translation_invariant:
        Translation Invariant Unbalanced Sinkhorn :ref:`[73] <references-sinkhorn-unbalanced-translation-invariant>`

    """

    if method.lower() == "sinkhorn":
        return sinkhorn_knopp_unbalanced(
            a,
            b,
            M,
            reg,
            reg_m,
            reg_type,
            c,
            warmstart,
            numItermax=numItermax,
            stopThr=stopThr,
            verbose=verbose,
            log=log,
            **kwargs,
        )

    elif method.lower() == "sinkhorn_stabilized":
        return sinkhorn_stabilized_unbalanced(
            a,
            b,
            M,
            reg,
            reg_m,
            reg_type,
            c,
            warmstart,
            numItermax=numItermax,
            stopThr=stopThr,
            verbose=verbose,
            log=log,
            **kwargs,
        )

    elif method.lower() == "sinkhorn_translation_invariant":
        return sinkhorn_unbalanced_translation_invariant(
            a,
            b,
            M,
            reg,
            reg_m,
            reg_type,
            c,
            warmstart,
            numItermax=numItermax,
            stopThr=stopThr,
            verbose=verbose,
            log=log,
            **kwargs,
        )

    elif method.lower() in ["sinkhorn_reg_scaling"]:
        warnings.warn("Method not implemented yet. Using classic Sinkhorn-Knopp")
        return sinkhorn_knopp_unbalanced(
            a,
            b,
            M,
            reg,
            reg_m,
            reg_type,
            c,
            warmstart,
            numItermax=numItermax,
            stopThr=stopThr,
            verbose=verbose,
            log=log,
            **kwargs,
        )
    else:
        raise ValueError("Unknown method '%s'." % method)


def sinkhorn_unbalanced2(
    a,
    b,
    M,
    reg,
    reg_m,
    method="sinkhorn",
    reg_type="kl",
    c=None,
    warmstart=None,
    returnCost="linear",
    numItermax=1000,
    stopThr=1e-6,
    verbose=False,
    log=False,
    **kwargs,
):
    r"""
    Solve the entropic regularization unbalanced optimal transport problem and
    return the cost

    The function solves the following optimization problem:

    .. math::
        \min_\gamma \quad \langle \gamma, \mathbf{M} \rangle_F +
        \mathrm{reg} \cdot \mathrm{KL}(\gamma, \mathbf{c}) +
        \mathrm{reg_{m1}} \cdot \mathrm{KL}(\gamma \mathbf{1}, \mathbf{a}) +
        \mathrm{reg_{m2}} \cdot \mathrm{KL}(\gamma^T \mathbf{1}, \mathbf{b})

        s.t.
             \gamma\geq 0
    where :

    - :math:`\mathbf{M}` is the (`dim_a`, `dim_b`) metric cost matrix
    - :math:`\mathbf{a}` and :math:`\mathbf{b}` are source and target unbalanced distributions
    - :math:`\mathbf{c}` is a reference distribution for the regularization
    - KL is the Kullback-Leibler divergence

    The algorithm used for solving the problem is the generalized
    Sinkhorn-Knopp matrix scaling algorithm as proposed in :ref:`[10, 25]
    <references-sinkhorn-unbalanced2>`

    .. warning::
        Starting from version 0.9.5, the default value has been changed to `reg_type='kl'` instead of `reg_type='entropy'`. This makes the function more consistent with the literature
        and the other solvers. If you want to use the entropy regularization, please set `reg_type='entropy'` explicitly.

    Parameters
    ----------
    a : array-like, shape (dim_a,)
        Unnormalized histogram of dimension `dim_a`
        If `a` is an empty list or array ([]),
        then `a` is set to uniform distribution.
    b : array-like, shape (dim_b,)
        One or multiple unnormalized histograms of dimension `dim_b`.
        If `b` is an empty list or array ([]),
        then `b` is set to uniform distribution.
        If many, compute all the OT costs :math:`(\mathbf{a}, \mathbf{b}_i)_i`
    M : array-like, shape (dim_a, dim_b)
        loss matrix
    reg : float
        Entropy regularization term > 0
    reg_m: float or indexable object of length 1 or 2
        Marginal relaxation term.
        If :math:`\mathrm{reg_{m}}` is a scalar or an indexable object of length 1,
        then the same :math:`\mathrm{reg_{m}}` is applied to both marginal relaxations.
        The entropic balanced OT can be recovered using :math:`\mathrm{reg_{m}}=float("inf")`.
        For semi-relaxed case, use either
        :math:`\mathrm{reg_{m}}=(float("inf"), scalar)` or
        :math:`\mathrm{reg_{m}}=(scalar, float("inf"))`.
        If :math:`\mathrm{reg_{m}}` is an array,
        it must have the same backend as input arrays `(a, b, M)`.
    method : str
        method used for the solver either 'sinkhorn', 'sinkhorn_stabilized', 'sinkhorn_translation_invariant' or
        'sinkhorn_reg_scaling', see those function for specific parameters
    reg_type : string, optional
        Regularizer term. Can take two values:

        - Negative entropy: 'entropy':
          :math:`\Omega(\gamma) = \sum_{i,j} \gamma_{i,j} \log(\gamma_{i,j}) - \sum_{i,j} \gamma_{i,j}`.
          This is equivalent (up to a constant) to :math:`\Omega(\gamma) = \text{KL}(\gamma, 1_{dim_a} 1_{dim_b}^T)`.
        - Kullback-Leibler divergence: 'kl':
          :math:`\Omega(\gamma) = \text{KL}(\gamma, \mathbf{a} \mathbf{b}^T)`.
    c : array-like, shape (dim_a, dim_b), optional (default=None)
        Reference measure for the regularization.
        If None, then use :math:`\mathbf{c} = \mathbf{a} \mathbf{b}^T`.
        If :math:`\texttt{reg_type}=`'entropy', then :math:`\mathbf{c} = 1_{dim_a} 1_{dim_b}^T`.
    warmstart: tuple of arrays, shape (dim_a, dim_b), optional
        Initialization of dual potentials. If provided, the dual potentials should be given
        (that is the logarithm of the u,v sinkhorn scaling vectors).
    returnCost: string, optional (default = "linear")
        If `returnCost` = "linear", then return the linear part of the unbalanced OT loss.
        If `returnCost` = "total", then return the total unbalanced OT loss.
    numItermax : int, optional
        Max number of iterations
    stopThr : float, optional
        Stop threshold on error (>0)
    verbose : bool, optional
        Print information along iterations
    log : bool, optional
        record `log` if `True`


    Returns
    -------
    ot_cost : array-like, shape (n_hists,)
        the OT cost between :math:`\mathbf{a}` and each of the histograms :math:`\mathbf{b}_i`
    log : dict
        log dictionary returned only if `log` is `True`

    Examples
    --------

    >>> import ot
    >>> import numpy as np
    >>> a=[.5, .10]
    >>> b=[.5, .5]
    >>> M=[[0., 1.],[1., 0.]]
    >>> np.round(ot.unbalanced.sinkhorn_unbalanced2(a, b, M, 1., 1.), 8)
    0.19600125


    .. _references-sinkhorn-unbalanced2:
    References
    ----------
    .. [2] M. Cuturi, Sinkhorn Distances : Lightspeed Computation of Optimal
        Transport, Advances in Neural Information Processing Systems
        (NIPS) 26, 2013

    .. [9] Schmitzer, B. (2016). Stabilized Sparse Scaling Algorithms for
        Entropy Regularized Transport Problems. arXiv preprint arXiv:1610.06519.

    .. [10] Chizat, L., Peyré, G., Schmitzer, B., & Vialard, F. X. (2016).
        Scaling algorithms for unbalanced transport problems. arXiv preprint
        arXiv:1607.05816.

    .. [25] Frogner C., Zhang C., Mobahi H., Araya-Polo M., Poggio T. :
        Learning with a Wasserstein Loss,  Advances in Neural Information
        Processing Systems (NIPS) 2015

    .. [73] Séjourné, T., Vialard, F. X., & Peyré, G. (2022).
        Faster unbalanced optimal transport: Translation invariant sinkhorn and 1-d frank-wolfe.
        In International Conference on Artificial Intelligence and Statistics (pp. 4995-5021). PMLR.

    See Also
    --------
    ot.unbalanced.sinkhorn_knopp: Unbalanced Classic Sinkhorn :ref:`[10] <references-sinkhorn-unbalanced2>`
    ot.unbalanced.sinkhorn_stabilized: Unbalanced Stabilized sinkhorn :ref:`[9, 10] <references-sinkhorn-unbalanced2>`
    ot.unbalanced.sinkhorn_reg_scaling: Unbalanced Sinkhorn with epsilon scaling :ref:`[9, 10] <references-sinkhorn-unbalanced2>`
    ot.unbalanced.sinkhorn_unbalanced_translation_invariant: Translation Invariant Unbalanced Sinkhorn :ref:`[73] <references-sinkhorn-unbalanced2>`

    """
    M, a, b = list_to_array(M, a, b)

    if len(b.shape) < 2:
        if method.lower() == "sinkhorn":
            res = sinkhorn_knopp_unbalanced(
                a,
                b,
                M,
                reg,
                reg_m,
                reg_type,
                c,
                warmstart,
                numItermax=numItermax,
                stopThr=stopThr,
                verbose=verbose,
                log=True,
                **kwargs,
            )

        elif method.lower() == "sinkhorn_stabilized":
            res = sinkhorn_stabilized_unbalanced(
                a,
                b,
                M,
                reg,
                reg_m,
                reg_type,
                c,
                warmstart,
                numItermax=numItermax,
                stopThr=stopThr,
                verbose=verbose,
                log=True,
                **kwargs,
            )

        elif method.lower() == "sinkhorn_translation_invariant":
            res = sinkhorn_unbalanced_translation_invariant(
                a,
                b,
                M,
                reg,
                reg_m,
                reg_type,
                c,
                warmstart,
                numItermax=numItermax,
                stopThr=stopThr,
                verbose=verbose,
                log=True,
                **kwargs,
            )

        elif method.lower() in ["sinkhorn_reg_scaling"]:
            warnings.warn("Method not implemented yet. Using classic Sinkhorn-Knopp")
            res = sinkhorn_knopp_unbalanced(
                a,
                b,
                M,
                reg,
                reg_m,
                reg_type,
                c,
                warmstart,
                numItermax=numItermax,
                stopThr=stopThr,
                verbose=verbose,
                log=True,
                **kwargs,
            )
        else:
            raise ValueError("Unknown method %s." % method)

        if returnCost == "linear":
            cost = res[1]["cost"]
        elif returnCost == "total":
            cost = res[1]["total_cost"]
        else:
            raise ValueError("Unknown returnCost = {}".format(returnCost))

        if log:
            return cost, res[1]
        else:
            return cost

    else:
        if reg_type == "kl":
            warnings.warn("Reg_type not implemented yet. Use entropy.")

        if method.lower() == "sinkhorn":
            return sinkhorn_knopp_unbalanced(
                a,
                b,
                M,
                reg,
                reg_m,
                reg_type,
                c,
                warmstart,
                numItermax=numItermax,
                stopThr=stopThr,
                verbose=verbose,
                log=log,
                **kwargs,
            )

        elif method.lower() == "sinkhorn_stabilized":
            return sinkhorn_stabilized_unbalanced(
                a,
                b,
                M,
                reg,
                reg_m,
                reg_type,
                c,
                warmstart,
                numItermax=numItermax,
                stopThr=stopThr,
                verbose=verbose,
                log=log,
                **kwargs,
            )

        elif method.lower() == "sinkhorn_translation_invariant":
            return sinkhorn_unbalanced_translation_invariant(
                a,
                b,
                M,
                reg,
                reg_m,
                reg_type,
                c,
                warmstart,
                numItermax=numItermax,
                stopThr=stopThr,
                verbose=verbose,
                log=log,
                **kwargs,
            )

        elif method.lower() in ["sinkhorn_reg_scaling"]:
            warnings.warn("Method not implemented yet. Using classic Sinkhorn-Knopp")
            return sinkhorn_knopp_unbalanced(
                a,
                b,
                M,
                reg,
                reg_m,
                reg_type,
                c,
                warmstart,
                numItermax=numItermax,
                stopThr=stopThr,
                verbose=verbose,
                log=log,
                **kwargs,
            )
        else:
            raise ValueError("Unknown method %s." % method)


def sinkhorn_knopp_unbalanced(
    a,
    b,
    M,
    reg,
    reg_m,
    reg_type="kl",
    c=None,
    warmstart=None,
    numItermax=1000,
    stopThr=1e-6,
    verbose=False,
    log=False,
    **kwargs,
):
    r"""
    Solve the entropic regularization unbalanced optimal transport problem and
    return the OT plan

    The function solves the following optimization problem:

    .. math::
        W = \arg \min_\gamma \quad \langle \gamma, \mathbf{M} \rangle_F +
        \mathrm{reg} \cdot \mathrm{KL}(\gamma, \mathbf{c}) +
        \mathrm{reg_{m1}} \cdot \mathrm{KL}(\gamma \mathbf{1}, \mathbf{a}) +
        \mathrm{reg_{m2}} \cdot \mathrm{KL}(\gamma^T \mathbf{1}, \mathbf{b})

        s.t.
             \gamma \geq 0

    where :

    - :math:`\mathbf{M}` is the (`dim_a`, `dim_b`) metric cost matrix
    - :math:`\mathbf{a}` and :math:`\mathbf{b}` are source and target unbalanced distributions
    - :math:`\mathbf{c}` is a reference distribution for the regularization
    - KL is the Kullback-Leibler divergence

    The algorithm used for solving the problem is the generalized Sinkhorn-Knopp matrix scaling algorithm as proposed in :ref:`[10, 25] <references-sinkhorn-knopp-unbalanced>`

    .. warning::
        Starting from version 0.9.5, the default value has been changed to `reg_type='kl'` instead of `reg_type='entropy'`. This makes the function more consistent with the literature
        and the other solvers. If you want to use the entropy regularization, please set `reg_type='entropy'` explicitly.


    Parameters
    ----------
    a : array-like, shape (dim_a,)
        Unnormalized histogram of dimension `dim_a`
        If `a` is an empty list or array ([]),
        then `a` is set to uniform distribution.
    b : array-like, shape (dim_b,)
        One or multiple unnormalized histograms of dimension `dim_b`.
        If `b` is an empty list or array ([]),
        then `b` is set to uniform distribution.
        If many, compute all the OT costs :math:`(\mathbf{a}, \mathbf{b}_i)_i`
    M : array-like, shape (dim_a, dim_b)
        loss matrix
    reg : float
        Entropy regularization term > 0
    reg_m: float or indexable object of length 1 or 2
        Marginal relaxation term.
        If :math:`\mathrm{reg_{m}}` is a scalar or an indexable object of length 1,
        then the same :math:`\mathrm{reg_{m}}` is applied to both marginal relaxations.
        The entropic balanced OT can be recovered using :math:`\mathrm{reg_{m}}=float("inf")`.
        For semi-relaxed case, use either
        :math:`\mathrm{reg_{m}}=(float("inf"), scalar)` or
        :math:`\mathrm{reg_{m}}=(scalar, float("inf"))`.
        If :math:`\mathrm{reg_{m}}` is an array,
        it must have the same backend as input arrays `(a, b, M)`.
    reg_type : string, optional
        Regularizer term. Can take two values:

        - Negative entropy: 'entropy':
          :math:`\Omega(\gamma) = \sum_{i,j} \gamma_{i,j} \log(\gamma_{i,j}) - \sum_{i,j} \gamma_{i,j}`.
          This is equivalent (up to a constant) to :math:`\Omega(\gamma) = \text{KL}(\gamma, 1_{dim_a} 1_{dim_b}^T)`.
        - Kullback-Leibler divergence: 'kl':
          :math:`\Omega(\gamma) = \text{KL}(\gamma, \mathbf{a} \mathbf{b}^T)`.
    c : array-like, shape (dim_a, dim_b), optional (default=None)
        Reference measure for the regularization.
        If None, then use :math:`\mathbf{c} = \mathbf{a} \mathbf{b}^T`.
        If :math:`\texttt{reg_type}=`'entropy', then :math:`\mathbf{c} = 1_{dim_a} 1_{dim_b}^T`.
    warmstart: tuple of arrays, shape (dim_a, dim_b), optional
        Initialization of dual potentials. If provided, the dual potentials should be given
        (that is the logarithm of the `u`, `v` sinkhorn scaling vectors).
    numItermax : int, optional
        Max number of iterations
    stopThr : float, optional
        Stop threshold on error (> 0)
    verbose : bool, optional
        Print information along iterations
    log : bool, optional
        record `log` if `True`


    Returns
    -------
    if n_hists == 1:
        - gamma : array-like, shape (dim_a, dim_b)
            Optimal transportation matrix for the given parameters
        - log : dict
            log dictionary returned only if `log` is `True`
    else:
        - ot_cost : array-like, shape (n_hists,)
            the OT cost between :math:`\mathbf{a}` and each of the histograms :math:`\mathbf{b}_i`
        - log : dict
            log dictionary returned only if `log` is `True`

    Examples
    --------

    >>> import ot
    >>> import numpy as np
    >>> a=[.5, .5]
    >>> b=[.5, .5]
    >>> M=[[0., 1.],[1., 0.]]
    >>> np.round(ot.unbalanced.sinkhorn_knopp_unbalanced(a, b, M, 1., 1.), 7)
    array([[0.3220536, 0.1184769],
           [0.1184769, 0.3220536]])


    .. _references-sinkhorn-knopp-unbalanced:
    References
    ----------
    .. [10] Chizat, L., Peyré, G., Schmitzer, B., & Vialard, F. X. (2016).
        Scaling algorithms for unbalanced transport problems. arXiv preprint
        arXiv:1607.05816.

    .. [25] Frogner C., Zhang C., Mobahi H., Araya-Polo M., Poggio T. :
        Learning with a Wasserstein Loss,  Advances in Neural Information
        Processing Systems (NIPS) 2015

    See Also
    --------
    ot.lp.emd : Unregularized OT
    ot.optim.cg : General regularized OT

    """
    M, a, b = list_to_array(M, a, b)
    nx = get_backend(M, a, b)

    dim_a, dim_b = M.shape

    if len(a) == 0:
        a = nx.ones(dim_a, type_as=M) / dim_a
    if len(b) == 0:
        b = nx.ones(dim_b, type_as=M) / dim_b

    if len(b.shape) > 1:
        n_hists = b.shape[1]
    else:
        n_hists = 0

    reg_m1, reg_m2 = get_parameter_pair(reg_m)

    if log:
        dict_log = {"err": []}

    # we assume that no distances are null except those of the diagonal of
    # distances
    if warmstart is None:
        if n_hists:
            u = nx.ones((dim_a, 1), type_as=M)
            v = nx.ones((dim_b, n_hists), type_as=M)
            a = a.reshape(dim_a, 1)
        else:
            u = nx.ones(dim_a, type_as=M)
            v = nx.ones(dim_b, type_as=M)
    else:
        u, v = nx.exp(warmstart[0]), nx.exp(warmstart[1])

    if reg_type.lower() == "entropy":
        warnings.warn(
            "If reg_type = entropy, then the matrix c is overwritten by the one matrix."
        )
        c = nx.ones((dim_a, dim_b), type_as=M)

    if n_hists:
        K = nx.exp(-M / reg)
    else:
        c = a[:, None] * b[None, :] if c is None else c
        K = nx.exp(-M / reg) * c

    fi_1 = reg_m1 / (reg_m1 + reg) if reg_m1 != float("inf") else 1
    fi_2 = reg_m2 / (reg_m2 + reg) if reg_m2 != float("inf") else 1

    err = 1.0

    for i in range(numItermax):
        uprev = u
        vprev = v

        Kv = nx.dot(K, v)
        u = (a / Kv) ** fi_1
        Ktu = nx.dot(K.T, u)
        v = (b / Ktu) ** fi_2

        if (
            nx.any(Ktu == 0.0)
            or nx.any(nx.isnan(u))
            or nx.any(nx.isnan(v))
            or nx.any(nx.isinf(u))
            or nx.any(nx.isinf(v))
        ):
            # we have reached the machine precision
            # come back to previous solution and quit loop
            warnings.warn("Numerical errors at iteration %s" % i)
            u = uprev
            v = vprev
            break

        err_u = nx.max(nx.abs(u - uprev)) / max(
            nx.max(nx.abs(u)), nx.max(nx.abs(uprev)), 1.0
        )
        err_v = nx.max(nx.abs(v - vprev)) / max(
            nx.max(nx.abs(v)), nx.max(nx.abs(vprev)), 1.0
        )
        err = 0.5 * (err_u + err_v)
        if log:
            dict_log["err"].append(err)
            if verbose:
                if i % 50 == 0:
                    print("{:5s}|{:12s}".format("It.", "Err") + "\n" + "-" * 19)
                print("{:5d}|{:8e}|".format(i, err))
        if err < stopThr:
            break

    if log:
        dict_log["logu"] = nx.log(u + 1e-300)
        dict_log["logv"] = nx.log(v + 1e-300)

    if n_hists:  # return only loss
        res = nx.einsum("ik,ij,jk,ij->k", u, K, v, M)
        if log:
            return res, dict_log
        else:
            return res

    else:  # return OT matrix
        plan = u[:, None] * K * v[None, :]

        if log:
            linear_cost = nx.sum(plan * M)
            dict_log["cost"] = linear_cost

            total_cost = linear_cost + reg * nx.kl_div(plan, c)
            if reg_m1 != float("inf"):
                total_cost = total_cost + reg_m1 * nx.kl_div(nx.sum(plan, 1), a)
            if reg_m2 != float("inf"):
                total_cost = total_cost + reg_m2 * nx.kl_div(nx.sum(plan, 0), b)
            dict_log["total_cost"] = total_cost

            return plan, dict_log
        else:
            return plan


def sinkhorn_stabilized_unbalanced(
    a,
    b,
    M,
    reg,
    reg_m,
    reg_type="kl",
    c=None,
    warmstart=None,
    tau=1e5,
    numItermax=1000,
    stopThr=1e-6,
    verbose=False,
    log=False,
    **kwargs,
):
    r"""
    Solve the entropic regularization unbalanced optimal transport
    problem and return the loss

    The function solves the following optimization problem using log-domain
    stabilization as proposed in :ref:`[10] <references-sinkhorn-stabilized-unbalanced>`:

    .. math::
        W = \arg \min_\gamma \quad \langle \gamma, \mathbf{M} \rangle_F +
        \mathrm{reg} \cdot \mathrm{KL}(\gamma, \mathbf{c}) +
        \mathrm{reg_{m1}} \cdot \mathrm{KL}(\gamma \mathbf{1}, \mathbf{a}) +
        \mathrm{reg_{m2}} \cdot \mathrm{KL}(\gamma^T \mathbf{1}, \mathbf{b})

        s.t.
             \gamma \geq 0

    where :

    - :math:`\mathbf{M}` is the (`dim_a`, `dim_b`) metric cost matrix
    - :math:`\mathbf{a}` and :math:`\mathbf{b}` are source and target unbalanced distributions
    - :math:`\mathbf{c}` is a reference distribution for the regularization
    - KL is the Kullback-Leibler divergence

    The algorithm used for solving the problem is the generalized
    Sinkhorn-Knopp matrix scaling algorithm as proposed in :ref:`[10, 25] <references-sinkhorn-stabilized-unbalanced>`


    Parameters
    ----------
    a : array-like, shape (dim_a,)
        Unnormalized histogram of dimension `dim_a`
        If `a` is an empty list or array ([]),
        then `a` is set to uniform distribution.
    b : array-like, shape (dim_b,)
        One or multiple unnormalized histograms of dimension `dim_b`.
        If `b` is an empty list or array ([]),
        then `b` is set to uniform distribution.
        If many, compute all the OT costs :math:`(\mathbf{a}, \mathbf{b}_i)_i`
    M : array-like, shape (dim_a, dim_b)
        loss matrix
    reg : float
        Entropy regularization term > 0
    reg_m: float or indexable object of length 1 or 2
        Marginal relaxation term.
        If :math:`\mathrm{reg_{m}}` is a scalar or an indexable object of length 1,
        then the same :math:`\mathrm{reg_{m}}` is applied to both marginal relaxations.
        The entropic balanced OT can be recovered using :math:`\mathrm{reg_{m}}=float("inf")`.
        For semi-relaxed case, use either
        :math:`\mathrm{reg_{m}}=(float("inf"), scalar)` or
        :math:`\mathrm{reg_{m}}=(scalar, float("inf"))`.
        If :math:`\mathrm{reg_{m}}` is an array,
        it must have the same backend as input arrays `(a, b, M)`.
    method : str
        method used for the solver either 'sinkhorn', 'sinkhorn_stabilized' or
        'sinkhorn_reg_scaling', see those function for specific parameters
    reg_type : string, optional
        Regularizer term. Can take two values:

        - Negative entropy: 'entropy':
          :math:`\Omega(\gamma) = \sum_{i,j} \gamma_{i,j} \log(\gamma_{i,j}) - \sum_{i,j} \gamma_{i,j}`.
          This is equivalent (up to a constant) to :math:`\Omega(\gamma) = \text{KL}(\gamma, 1_{dim_a} 1_{dim_b}^T)`.
        - Kullback-Leibler divergence: 'kl':
          :math:`\Omega(\gamma) = \text{KL}(\gamma, \mathbf{a} \mathbf{b}^T)`.
    c : array-like, shape (dim_a, dim_b), optional (default=None)
        Reference measure for the regularization.
        If None, then use :math:`\mathbf{c} = \mathbf{a} \mathbf{b}^T`.
        If :math:`\texttt{reg_type}=`'entropy', then :math:`\mathbf{c} = 1_{dim_a} 1_{dim_b}^T`.
    warmstart: tuple of arrays, shape (dim_a, dim_b), optional
        Initialization of dual potentials. If provided, the dual potentials should be given
        (that is the logarithm of the `u`, `v` sinkhorn scaling vectors).
    tau : float
        threshold for max value in `u` or `v` for log scaling
    numItermax : int, optional
        Max number of iterations
    stopThr : float, optional
        Stop threshold on error (>0)
    verbose : bool, optional
        Print information along iterations
    log : bool, optional
        record `log` if `True`

    .. warning::
        Starting from version 0.9.5, the default value has been changed to `reg_type='kl'` instead of `reg_type='entropy'`. This makes the function more consistent with the literature
        and the other solvers. If you want to use the entropy regularization, please set `reg_type='entropy'` explicitly.


    Returns
    -------
    if n_hists == 1:
        - gamma : array-like, shape (dim_a, dim_b)
            Optimal transportation matrix for the given parameters
        - log : dict
            log dictionary returned only if `log` is `True`
    else:
        - ot_cost : array-like, shape (n_hists,)
            the OT cost between :math:`\mathbf{a}` and each of the histograms :math:`\mathbf{b}_i`
        - log : dict
            log dictionary returned only if `log` is `True`
    Examples
    --------

    >>> import ot
    >>> import numpy as np
    >>> a=[.5, .5]
    >>> b=[.5, .5]
    >>> M=[[0., 1.],[1., 0.]]
    >>> np.round(ot.unbalanced.sinkhorn_stabilized_unbalanced(a, b, M, 1., 1.), 7)
    array([[0.3220536, 0.1184769],
           [0.1184769, 0.3220536]])


    .. _references-sinkhorn-stabilized-unbalanced:
    References
    ----------
    .. [10] Chizat, L., Peyré, G., Schmitzer, B., & Vialard, F. X. (2016).
        Scaling algorithms for unbalanced transport problems. arXiv preprint arXiv:1607.05816.

    .. [25] Frogner C., Zhang C., Mobahi H., Araya-Polo M., Poggio T. :
        Learning with a Wasserstein Loss,  Advances in Neural Information
        Processing Systems (NIPS) 2015

    See Also
    --------
    ot.lp.emd : Unregularized OT
    ot.optim.cg : General regularized OT

    """
    a, b, M = list_to_array(a, b, M)
    nx = get_backend(M, a, b)

    dim_a, dim_b = M.shape

    if len(a) == 0:
        a = nx.ones(dim_a, type_as=M) / dim_a
    if len(b) == 0:
        b = nx.ones(dim_b, type_as=M) / dim_b

    if len(b.shape) > 1:
        n_hists = b.shape[1]
    else:
        n_hists = 0

    reg_m1, reg_m2 = get_parameter_pair(reg_m)

    if log:
        dict_log = {"err": []}

    # we assume that no distances are null except those of the diagonal of
    # distances
    if warmstart is None:
        if n_hists:
            u = nx.ones((dim_a, n_hists), type_as=M)
            v = nx.ones((dim_b, n_hists), type_as=M)
            a = a.reshape(dim_a, 1)
        else:
            u = nx.ones(dim_a, type_as=M)
            v = nx.ones(dim_b, type_as=M)
    else:
        u, v = nx.exp(warmstart[0]), nx.exp(warmstart[1])

    if reg_type == "entropy":
        warnings.warn(
            "If reg_type = entropy, then the matrix c is overwritten by the one matrix."
        )
        c = nx.ones((dim_a, dim_b), type_as=M)

    if n_hists:
        M0 = M
    else:
        c = a[:, None] * b[None, :] if c is None else c
        M0 = M - reg * nx.log(c)
    K = nx.exp(-M0 / reg)

    fi_1 = reg_m1 / (reg_m1 + reg) if reg_m1 != float("inf") else 1
    fi_2 = reg_m2 / (reg_m2 + reg) if reg_m2 != float("inf") else 1

    cpt = 0
    err = 1.0
    alpha = nx.zeros(dim_a, type_as=M)
    beta = nx.zeros(dim_b, type_as=M)
    ones_a = nx.ones(dim_a, type_as=M)
    ones_b = nx.ones(dim_b, type_as=M)

    while err > stopThr and cpt < numItermax:
        uprev = u
        vprev = v

        Kv = nx.dot(K, v)
        f_alpha = nx.exp(-alpha / (reg + reg_m1)) if reg_m1 != float("inf") else ones_a
        f_beta = nx.exp(-beta / (reg + reg_m2)) if reg_m2 != float("inf") else ones_b

        if n_hists:
            f_alpha = f_alpha[:, None]
            f_beta = f_beta[:, None]
        u = ((a / (Kv + 1e-16)) ** fi_1) * f_alpha
        Ktu = nx.dot(K.T, u)
        v = ((b / (Ktu + 1e-16)) ** fi_2) * f_beta
        absorbing = False
        if nx.any(u > tau) or nx.any(v > tau):
            absorbing = True
            if n_hists:
                alpha = alpha + reg * nx.log(nx.max(u, 1))
                beta = beta + reg * nx.log(nx.max(v, 1))
            else:
                alpha = alpha + reg * nx.log(nx.max(u))
                beta = beta + reg * nx.log(nx.max(v))
            K = nx.exp((alpha[:, None] + beta[None, :] - M0) / reg)
            v = nx.ones(v.shape, type_as=v)
        Kv = nx.dot(K, v)

        if (
            nx.any(Ktu == 0.0)
            or nx.any(nx.isnan(u))
            or nx.any(nx.isnan(v))
            or nx.any(nx.isinf(u))
            or nx.any(nx.isinf(v))
        ):
            # we have reached the machine precision
            # come back to previous solution and quit loop
            warnings.warn("Numerical errors at iteration %s" % cpt)
            u = uprev
            v = vprev
            break
        if (cpt % 10 == 0 and not absorbing) or cpt == 0:
            # we can speed up the process by checking for the error only all
            # the 10th iterations
            err = nx.max(nx.abs(u - uprev)) / max(
                nx.max(nx.abs(u)), nx.max(nx.abs(uprev)), 1.0
            )
            if log:
                dict_log["err"].append(err)
            if verbose:
                if cpt % 200 == 0:
                    print("{:5s}|{:12s}".format("It.", "Err") + "\n" + "-" * 19)
                print("{:5d}|{:8e}|".format(cpt, err))
        cpt = cpt + 1

    if err > stopThr:
        warnings.warn(
            "Stabilized Unbalanced Sinkhorn did not converge."
            + "Try a larger entropy `reg` or a lower mass `reg_m`."
            + "Or a larger absorption threshold `tau`."
        )
    if n_hists:
        logu = alpha[:, None] / reg + nx.log(u)
        logv = beta[:, None] / reg + nx.log(v)
    else:
        logu = alpha / reg + nx.log(u)
        logv = beta / reg + nx.log(v)
    if log:
        dict_log["logu"] = logu
        dict_log["logv"] = logv
    if n_hists:  # return only loss
        res = nx.logsumexp(
            nx.log(M + 1e-100)[:, :, None]
            + logu[:, None, :]
            + logv[None, :, :]
            - M0[:, :, None] / reg,
            axis=(0, 1),
        )
        res = nx.exp(res)
        if log:
            return res, dict_log
        else:
            return res

    else:  # return OT matrix
        plan = nx.exp(logu[:, None] + logv[None, :] - M0 / reg)
        if log:
            linear_cost = nx.sum(plan * M)
            dict_log["cost"] = linear_cost

            total_cost = linear_cost + reg * nx.kl_div(plan, c)
            if reg_m1 != float("inf"):
                total_cost = total_cost + reg_m1 * nx.kl_div(nx.sum(plan, 1), a)
            if reg_m2 != float("inf"):
                total_cost = total_cost + reg_m2 * nx.kl_div(nx.sum(plan, 0), b)
            dict_log["total_cost"] = total_cost

            return plan, dict_log
        else:
            return plan


def sinkhorn_unbalanced_translation_invariant(
    a,
    b,
    M,
    reg,
    reg_m,
    reg_type="kl",
    c=None,
    warmstart=None,
    numItermax=1000,
    stopThr=1e-6,
    verbose=False,
    log=False,
    **kwargs,
):
    r"""
    Solve the entropic regularization unbalanced optimal transport problem and
    return the OT plan

    The function solves the following optimization problem:

    .. math::
        W = \arg \min_\gamma \ \langle \gamma, \mathbf{M} \rangle_F +
        \mathrm{reg} \cdot \mathrm{KL}(\gamma, \mathbf{c}) +
        \mathrm{reg_{m1}} \cdot \mathrm{KL}(\gamma \mathbf{1}, \mathbf{a}) +
        \mathrm{reg_{m2}} \cdot \mathrm{KL}(\gamma^T \mathbf{1}, \mathbf{b})

        s.t.
             \gamma \geq 0

    where :

    - :math:`\mathbf{M}` is the (`dim_a`, `dim_b`) metric cost matrix
    - :math:`\Omega` is the entropic regularization term,KL divergence
    - :math:`\mathbf{a}` and :math:`\mathbf{b}` are source and target unbalanced distributions
    - KL is the Kullback-Leibler divergence

    The algorithm used for solving the problem is the translation invariant Sinkhorn algorithm as proposed in :ref:`[73] <references-sinkhorn-unbalanced-translation-invariant>`

    Parameters
    ----------
    a : array-like, shape (dim_a,)
        Unnormalized histogram of dimension `dim_a`
    b : array-like, shape (dim_b,) or (dim_b, n_hists)
        One or multiple unnormalized histograms of dimension `dim_b`
        If many, compute all the OT distances (a, b_i)
    M : array-like, shape (dim_a, dim_b)
        loss matrix
    reg : float
        Entropy regularization term > 0
    reg_m: float or indexable object of length 1 or 2
        Marginal relaxation term.
        If reg_m is a scalar or an indexable object of length 1,
        then the same reg_m is applied to both marginal relaxations.
        The entropic balanced OT can be recovered using `reg_m=float("inf")`.
        For semi-relaxed case, use either
        `reg_m=(float("inf"), scalar)` or `reg_m=(scalar, float("inf"))`.
        If reg_m is an array, it must have the same backend as input arrays (a, b, M).
    reg_type : string, optional
        Regularizer term. Can take two values:
        'entropy' (negative entropy)
        :math:`\Omega(\gamma) = \sum_{i,j} \gamma_{i,j} \log(\gamma_{i,j}) - \sum_{i,j} \gamma_{i,j}`, or
        'kl' (Kullback-Leibler)
        :math:`\Omega(\gamma) = \text{KL}(\gamma, \mathbf{a} \mathbf{b}^T)`.
    c : array-like, shape (dim_a, dim_b), optional (default=None)
        Reference measure for the regularization.
        If None, then use :math:`\mathbf{c} = \mathbf{a} \mathbf{b}^T`.
        If :math:`\texttt{reg_type}=`'entropy', then :math:`\mathbf{c} = 1_{dim_a} 1_{dim_b}^T`.
    warmstart: tuple of arrays, shape (dim_a, dim_b), optional
        Initialization of dual potentials. If provided, the dual potentials should be given
        (that is the logarithm of the u,v sinkhorn scaling vectors).
    numItermax : int, optional
        Max number of iterations
    stopThr : float, optional
        Stop threshold on error (> 0)
    verbose : bool, optional
        Print information along iterations
    log : bool, optional
        record log if True


    Returns
    -------
    if n_hists == 1:
        - gamma : array-like, shape (dim_a, dim_b)
            Optimal transportation matrix for the given parameters
        - log : dict
            log dictionary returned only if `log` is `True`
    else:
        - ot_distance : array-like, shape (n_hists,)
            the OT distance between :math:`\mathbf{a}` and each of the histograms :math:`\mathbf{b}_i`
        - log : dict
            log dictionary returned only if `log` is `True`

    Examples
    --------

    >>> import ot
    >>> a=[.5, .5]
    >>> b=[.5, .5]
    >>> M=[[0., 1.],[1., 0.]]
    >>> ot.unbalanced.sinkhorn_unbalanced_translation_invariant(a, b, M, 1., 1.)
    array([[0.32205357, 0.11847689],
           [0.11847689, 0.32205357]])


    .. _references-sinkhorn-unbalanced-translation-invariant:
    References
    ----------
    .. [73] Séjourné, T., Vialard, F. X., & Peyré, G. (2022).
        Faster unbalanced optimal transport: Translation invariant sinkhorn and 1-d frank-wolfe.
        In International Conference on Artificial Intelligence and Statistics (pp. 4995-5021). PMLR.
    """

    M, a, b = list_to_array(M, a, b)
    nx = get_backend(M, a, b)

    dim_a, dim_b = M.shape

    if len(a) == 0:
        a = nx.ones(dim_a, type_as=M) / dim_a
    if len(b) == 0:
        b = nx.ones(dim_b, type_as=M) / dim_b

    if len(b.shape) > 1:
        n_hists = b.shape[1]
    else:
        n_hists = 0

    reg_m1, reg_m2 = get_parameter_pair(reg_m)

    if log:
        dict_log = {"err": []}

    # we assume that no distances are null except those of the diagonal of
    # distances
    if warmstart is None:
        if n_hists:
            u = nx.ones((dim_a, 1), type_as=M)
            v = nx.ones((dim_b, n_hists), type_as=M)
            a = a.reshape(dim_a, 1)
        else:
            u = nx.ones(dim_a, type_as=M)
            v = nx.ones(dim_b, type_as=M)
    else:
        u, v = nx.exp(warmstart[0]), nx.exp(warmstart[1])

    u_, v_ = u, v

    if reg_type == "entropy":
        warnings.warn(
            "If reg_type = entropy, then the matrix c is overwritten by the one matrix."
        )
        c = nx.ones((dim_a, dim_b), type_as=M)

    if n_hists:
        M0 = M
    else:
        c = a[:, None] * b[None, :] if c is None else c
        M0 = M - reg * nx.log(c)
    K = nx.exp(-M0 / reg)

    fi_1 = reg_m1 / (reg_m1 + reg) if reg_m1 != float("inf") else 1
    fi_2 = reg_m2 / (reg_m2 + reg) if reg_m2 != float("inf") else 1

    k1 = (
        reg * reg_m1 / ((reg + reg_m1) * (reg_m1 + reg_m2))
        if reg_m1 != float("inf")
        else 0
    )
    k2 = (
        reg * reg_m2 / ((reg + reg_m2) * (reg_m1 + reg_m2))
        if reg_m2 != float("inf")
        else 0
    )

    k_rho1 = k1 * reg_m1 / reg if reg_m1 != float("inf") else 0
    k_rho2 = k2 * reg_m2 / reg if reg_m2 != float("inf") else 0

    if reg_m1 == float("inf") and reg_m2 == float("inf"):
        xi1, xi2 = 0, 0
        fi_12 = 1
    elif reg_m1 == float("inf"):
        xi1 = 0
        xi2 = reg / reg_m2
        fi_12 = reg_m2
    elif reg_m2 == float("inf"):
        xi1 = reg / reg_m1
        xi2 = 0
        fi_12 = reg_m1
    else:
        xi1 = (reg_m2 * reg) / (reg_m1 * (reg + reg_m1 + reg_m2))
        xi2 = (reg_m1 * reg) / (reg_m2 * (reg + reg_m1 + reg_m2))
        fi_12 = reg_m1 * reg_m2 / (reg_m1 + reg_m2)

    xi_rho1 = xi1 * reg_m1 / reg if reg_m1 != float("inf") else 0
    xi_rho2 = xi2 * reg_m2 / reg if reg_m2 != float("inf") else 0

    reg_ratio1 = reg / reg_m1 if reg_m1 != float("inf") else 0
    reg_ratio2 = reg / reg_m2 if reg_m2 != float("inf") else 0

    err = 1.0

    for i in range(numItermax):
        uprev = u
        vprev = v

        Kv = nx.dot(K, v_)
        u_hat = (a / Kv) ** fi_1 * nx.sum(b * v_**reg_ratio2) ** k_rho2
        u_ = u_hat * nx.sum(a * u_hat ** (-reg_ratio1)) ** (-xi_rho1)

        Ktu = nx.dot(K.T, u_)
        v_hat = (b / Ktu) ** fi_2 * nx.sum(a * u_ ** (-reg_ratio1)) ** k_rho1
        v_ = v_hat * nx.sum(b * v_hat ** (-reg_ratio2)) ** (-xi_rho2)

        if (
            nx.any(Ktu == 0.0)
            or nx.any(nx.isnan(u_))
            or nx.any(nx.isnan(v_))
            or nx.any(nx.isinf(u_))
            or nx.any(nx.isinf(v_))
        ):
            # we have reached the machine precision
            # come back to previous solution and quit loop
            warnings.warn("Numerical errors at iteration %s" % i)
            u = uprev
            v = vprev
            break

        t = (nx.sum(a * u_ ** (-reg_ratio1)) / nx.sum(b * v_ ** (-reg_ratio2))) ** (
            fi_12 / reg
        )

        u = u_ * t
        v = v_ / t

        err_u = nx.max(nx.abs(u - uprev)) / max(
            nx.max(nx.abs(u)), nx.max(nx.abs(uprev)), 1.0
        )
        err_v = nx.max(nx.abs(v - vprev)) / max(
            nx.max(nx.abs(v)), nx.max(nx.abs(vprev)), 1.0
        )
        err = 0.5 * (err_u + err_v)
        if log:
            dict_log["err"].append(err)
            if verbose:
                if i % 50 == 0:
                    print("{:5s}|{:12s}".format("It.", "Err") + "\n" + "-" * 19)
                print("{:5d}|{:8e}|".format(i, err))

        if err < stopThr:
            break

    if log:
        dict_log["logu"] = nx.log(u + 1e-300)
        dict_log["logv"] = nx.log(v + 1e-300)

    if n_hists:  # return only loss
        res = nx.einsum("ik,ij,jk,ij->k", u, K, v, M)
        if log:
            return res, dict_log
        else:
            return res

    else:  # return OT matrix
        plan = u[:, None] * K * v[None, :]

        if log:
            linear_cost = nx.sum(plan * M)
            dict_log["cost"] = linear_cost

            total_cost = linear_cost + reg * nx.kl_div(plan, c)
            if reg_m1 != float("inf"):
                total_cost = total_cost + reg_m1 * nx.kl_div(nx.sum(plan, 1), a)
            if reg_m2 != float("inf"):
                total_cost = total_cost + reg_m2 * nx.kl_div(nx.sum(plan, 0), b)
            dict_log["total_cost"] = total_cost

            return plan, dict_log
        else:
            return plan


def barycenter_unbalanced_stabilized(
    A,
    M,
    reg,
    reg_m,
    weights=None,
    tau=1e3,
    numItermax=1000,
    stopThr=1e-6,
    verbose=False,
    log=False,
):
    r"""Compute the entropic unbalanced wasserstein barycenter of :math:`\mathbf{A}` with stabilization.

     The function solves the following optimization problem:

    .. math::
       \mathbf{a} = \mathop{\arg \min}_\mathbf{a} \quad \sum_i W_{u_{reg}}(\mathbf{a},\mathbf{a}_i)

    where :

    - :math:`W_{u_{reg}}(\cdot,\cdot)` is the unbalanced entropic regularized Wasserstein distance (see :py:func:`ot.unbalanced.sinkhorn_unbalanced`)
    - :math:`\mathbf{a}_i` are training distributions in the columns of matrix :math:`\mathbf{A}`
    - reg and :math:`\mathbf{M}` are respectively the regularization term and the cost matrix for OT
    - reg_mis the marginal relaxation hyperparameter

    The algorithm used for solving the problem is the generalized
    Sinkhorn-Knopp matrix scaling algorithm as proposed in :ref:`[10] <references-barycenter-unbalanced-stabilized>`

    Parameters
    ----------
    A : array-like, shape (dim, n_hists)
        `n_hists` training distributions :math:`\mathbf{a}_i` of dimension `dim`
    M : array-like, shape (dim, dim)
        ground metric matrix for OT.
    reg : float
        Entropy regularization term > 0
    reg_m : float
        Marginal relaxation term > 0
    tau : float
        Stabilization threshold for log domain absorption.
    weights : array-like, shape (n_hists,) optional
        Weight of each distribution (barycentric coordinates)
        If None, uniform weights are used.
    numItermax : int, optional
        Max number of iterations
    stopThr : float, optional
        Stop threshold on error (> 0)
    verbose : bool, optional
        Print information along iterations
    log : bool, optional
        record `log` if `True`


    Returns
    -------
    a : array-like, shape (dim,)
        Unbalanced Wasserstein barycenter
    log : dict
        log dictionary return only if :math:`log==True` in parameters


    .. _references-barycenter-unbalanced-stabilized:
    References
    ----------
    .. [3] Benamou, J. D., Carlier, G., Cuturi, M., Nenna, L., & Peyré,
        G. (2015). Iterative Bregman projections for regularized transportation
        problems. SIAM Journal on Scientific Computing, 37(2), A1111-A1138.
    .. [10] Chizat, L., Peyré, G., Schmitzer, B., & Vialard, F. X. (2016).
        Scaling algorithms for unbalanced transport problems. arXiv preprint
        arXiv:1607.05816.


    """
    A, M = list_to_array(A, M)
    nx = get_backend(A, M)

    dim, n_hists = A.shape
    if weights is None:
        weights = nx.ones(n_hists, type_as=A) / n_hists
    else:
        assert len(weights) == A.shape[1]

    if log:
        log = {"err": []}

    fi = reg_m / (reg_m + reg)

    u = nx.ones((dim, n_hists), type_as=A) / dim
    v = nx.ones((dim, n_hists), type_as=A) / dim

    # print(reg)
    K = nx.exp(-M / reg)

    fi = reg_m / (reg_m + reg)

    cpt = 0
    err = 1.0
    alpha = nx.zeros(dim, type_as=A)
    beta = nx.zeros(dim, type_as=A)
    q = nx.ones(dim, type_as=A) / dim
    for i in range(numItermax):
        qprev = nx.copy(q)
        Kv = nx.dot(K, v)
        f_alpha = nx.exp(-alpha / (reg + reg_m))
        f_beta = nx.exp(-beta / (reg + reg_m))
        f_alpha = f_alpha[:, None]
        f_beta = f_beta[:, None]
        u = ((A / (Kv + 1e-16)) ** fi) * f_alpha
        Ktu = nx.dot(K.T, u)
        q = (Ktu ** (1 - fi)) * f_beta
        q = nx.dot(q, weights) ** (1 / (1 - fi))
        Q = q[:, None]
        v = ((Q / (Ktu + 1e-16)) ** fi) * f_beta
        absorbing = False
        if nx.any(u > tau) or nx.any(v > tau):
            absorbing = True
            alpha = alpha + reg * nx.log(nx.max(u, 1))
            beta = beta + reg * nx.log(nx.max(v, 1))
            K = nx.exp((alpha[:, None] + beta[None, :] - M) / reg)
            v = nx.ones(v.shape, type_as=v)
        Kv = nx.dot(K, v)
        if (
            nx.any(Ktu == 0.0)
            or nx.any(nx.isnan(u))
            or nx.any(nx.isnan(v))
            or nx.any(nx.isinf(u))
            or nx.any(nx.isinf(v))
        ):
            # we have reached the machine precision
            # come back to previous solution and quit loop
            warnings.warn("Numerical errors at iteration %s" % cpt)
            q = qprev
            break
        if (i % 10 == 0 and not absorbing) or i == 0:
            # we can speed up the process by checking for the error only all
            # the 10th iterations
            err = nx.max(nx.abs(q - qprev)) / max(
                nx.max(nx.abs(q)), nx.max(nx.abs(qprev)), 1.0
            )
            if log:
                log["err"].append(err)
            if verbose:
                if i % 50 == 0:
                    print("{:5s}|{:12s}".format("It.", "Err") + "\n" + "-" * 19)
                print("{:5d}|{:8e}|".format(i, err))
            if err < stopThr:
                break

    if err > stopThr:
        warnings.warn(
            "Stabilized Unbalanced Sinkhorn did not converge."
            + "Try a larger entropy `reg` or a lower mass `reg_m`."
            + "Or a larger absorption threshold `tau`."
        )
    if log:
        log["niter"] = i
        log["logu"] = nx.log(u + 1e-300)
        log["logv"] = nx.log(v + 1e-300)
        return q, log
    else:
        return q


def barycenter_unbalanced_sinkhorn(
    A,
    M,
    reg,
    reg_m,
    weights=None,
    numItermax=1000,
    stopThr=1e-6,
    verbose=False,
    log=False,
):
    r"""Compute the entropic unbalanced wasserstein barycenter of :math:`\mathbf{A}`.

     The function solves the following optimization problem with :math:`\mathbf{a}`

    .. math::
       \mathbf{a} = \mathop{\arg \min}_\mathbf{a} \quad \sum_i W_{u_{reg}}(\mathbf{a},\mathbf{a}_i)

    where :

    - :math:`W_{u_{reg}}(\cdot,\cdot)` is the unbalanced entropic regularized Wasserstein distance (see :py:func:`ot.unbalanced.sinkhorn_unbalanced`)
    - :math:`\mathbf{a}_i` are training distributions in the columns of matrix :math:`\mathbf{A}`
    - reg and :math:`\mathbf{M}` are respectively the regularization term and the cost matrix for OT
    - reg_mis the marginal relaxation hyperparameter

    The algorithm used for solving the problem is the generalized
    Sinkhorn-Knopp matrix scaling algorithm as proposed in :ref:`[10] <references-barycenter-unbalanced-sinkhorn>`

    Parameters
    ----------
    A : array-like, shape (dim, n_hists)
        `n_hists` training distributions :math:`\mathbf{a}_i` of dimension `dim`
    M : array-like, shape (dim, dim)
        ground metric matrix for OT.
    reg : float
        Entropy regularization term > 0
    reg_m: float
        Marginal relaxation term > 0
    weights : array-like, shape (n_hists,) optional
        Weight of each distribution (barycentric coordinates)
        If None, uniform weights are used.
    numItermax : int, optional
        Max number of iterations
    stopThr : float, optional
        Stop threshold on error (> 0)
    verbose : bool, optional
        Print information along iterations
    log : bool, optional
        record `log` if `True`


    Returns
    -------
    a : array-like, shape (dim,)
        Unbalanced Wasserstein barycenter
    log : dict
        log dictionary return only if :math:`log==True` in parameters


    .. _references-barycenter-unbalanced-sinkhorn:
    References
    ----------
    .. [3] Benamou, J. D., Carlier, G., Cuturi, M., Nenna, L., & Peyré, G.
        (2015). Iterative Bregman projections for regularized transportation
        problems. SIAM Journal on Scientific Computing, 37(2), A1111-A1138.
    .. [10] Chizat, L., Peyré, G., Schmitzer, B., & Vialard, F. X. (2016).
        Scaling algorithms for unbalanced transport problems. arXiv preprin
        arXiv:1607.05816.


    """
    A, M = list_to_array(A, M)
    nx = get_backend(A, M)

    dim, n_hists = A.shape
    if weights is None:
        weights = nx.ones(n_hists, type_as=A) / n_hists
    else:
        assert len(weights) == A.shape[1]

    if log:
        log = {"err": []}

    K = nx.exp(-M / reg)

    fi = reg_m / (reg_m + reg)

    v = nx.ones((dim, n_hists), type_as=A)
    u = nx.ones((dim, 1), type_as=A)
    q = nx.ones(dim, type_as=A)
    err = 1.0

    for i in range(numItermax):
        uprev = nx.copy(u)
        vprev = nx.copy(v)
        qprev = nx.copy(q)

        Kv = nx.dot(K, v)
        u = (A / Kv) ** fi
        Ktu = nx.dot(K.T, u)
        q = nx.dot(Ktu ** (1 - fi), weights)
        q = q ** (1 / (1 - fi))
        Q = q[:, None]
        v = (Q / Ktu) ** fi

        if (
            nx.any(Ktu == 0.0)
            or nx.any(nx.isnan(u))
            or nx.any(nx.isnan(v))
            or nx.any(nx.isinf(u))
            or nx.any(nx.isinf(v))
        ):
            # we have reached the machine precision
            # come back to previous solution and quit loop
            warnings.warn("Numerical errors at iteration %s" % i)
            u = uprev
            v = vprev
            q = qprev
            break
            # compute change in barycenter
        err = nx.max(nx.abs(q - qprev)) / max(
            nx.max(nx.abs(q)), nx.max(nx.abs(qprev)), 1.0
        )
        if log:
            log["err"].append(err)
        # if barycenter did not change + at least 10 iterations - stop
        if err < stopThr and i > 10:
            break

        if verbose:
            if i % 10 == 0:
                print("{:5s}|{:12s}".format("It.", "Err") + "\n" + "-" * 19)
            print("{:5d}|{:8e}|".format(i, err))

    if log:
        log["niter"] = i
        log["logu"] = nx.log(u + 1e-300)
        log["logv"] = nx.log(v + 1e-300)
        return q, log
    else:
        return q


def barycenter_unbalanced(
    A,
    M,
    reg,
    reg_m,
    method="sinkhorn",
    weights=None,
    numItermax=1000,
    stopThr=1e-6,
    verbose=False,
    log=False,
    **kwargs,
):
    r"""Compute the entropic unbalanced wasserstein barycenter of :math:`\mathbf{A}`.

     The function solves the following optimization problem with :math:`\mathbf{a}`

    .. math::
       \mathbf{a} = \mathop{\arg \min}_\mathbf{a} \quad \sum_i W_{u_{reg}}(\mathbf{a},\mathbf{a}_i)

    where :

    - :math:`W_{u_{reg}}(\cdot,\cdot)` is the unbalanced entropic regularized Wasserstein distance (see :py:func:`ot.unbalanced.sinkhorn_unbalanced`)
    - :math:`\mathbf{a}_i` are training distributions in the columns of matrix :math:`\mathbf{A}`
    - reg and :math:`\mathbf{M}` are respectively the regularization term and the cost matrix for OT
    - reg_mis the marginal relaxation hyperparameter

    The algorithm used for solving the problem is the generalized
    Sinkhorn-Knopp matrix scaling algorithm as proposed in :ref:`[10] <references-barycenter-unbalanced>`

    Parameters
    ----------
    A : array-like, shape (dim, n_hists)
        `n_hists` training distributions :math:`\mathbf{a}_i` of dimension `dim`
    M : array-like, shape (dim, dim)
        ground metric matrix for OT.
    reg : float
        Entropy regularization term > 0
    reg_m: float
        Marginal relaxation term > 0
    weights : array-like, shape (n_hists,) optional
        Weight of each distribution (barycentric coordinates)
        If None, uniform weights are used.
    numItermax : int, optional
        Max number of iterations
    stopThr : float, optional
        Stop threshold on error (> 0)
    verbose : bool, optional
        Print information along iterations
    log : bool, optional
        record log if True


    Returns
    -------
    a : array-like, shape (dim,)
        Unbalanced Wasserstein barycenter
    log : dict
        log dictionary return only if log==True in parameters


    .. _references-barycenter-unbalanced:
    References
    ----------
    .. [3] Benamou, J. D., Carlier, G., Cuturi, M., Nenna, L., & Peyré, G.
        (2015). Iterative Bregman projections for regularized transportation
        problems. SIAM Journal on Scientific Computing, 37(2), A1111-A1138.
    .. [10] Chizat, L., Peyré, G., Schmitzer, B., & Vialard, F. X. (2016).
        Scaling algorithms for unbalanced transport problems. arXiv preprin
        arXiv:1607.05816.

    """

    if method.lower() == "sinkhorn":
        return barycenter_unbalanced_sinkhorn(
            A,
            M,
            reg,
            reg_m,
            weights=weights,
            numItermax=numItermax,
            stopThr=stopThr,
            verbose=verbose,
            log=log,
            **kwargs,
        )

    elif method.lower() == "sinkhorn_stabilized":
        return barycenter_unbalanced_stabilized(
            A,
            M,
            reg,
            reg_m,
            weights=weights,
            numItermax=numItermax,
            stopThr=stopThr,
            verbose=verbose,
            log=log,
            **kwargs,
        )
    elif method.lower() in ["sinkhorn_reg_scaling", "sinkhorn_translation_invariant"]:
        warnings.warn("Method not implemented yet. Using classic Sinkhorn Knopp")
        return barycenter_unbalanced(
            A,
            M,
            reg,
            reg_m,
            weights=weights,
            numItermax=numItermax,
            stopThr=stopThr,
            verbose=verbose,
            log=log,
            **kwargs,
        )
    else:
        raise ValueError("Unknown method '%s'." % method)



"""Tests for module Unbalanced Co-Optimal Transport"""

# Author: Quang Huy Tran <quang-huy.tran@univ-ubs.fr>
#
# License: MIT License

import itertools
import numpy as np
import ot
import pytest
from ot.gromov._unbalanced import (
    unbalanced_co_optimal_transport,
    unbalanced_co_optimal_transport2,
)


@pytest.skip_backend("jax", reason="test very slow with jax backend")
@pytest.skip_backend("tf", reason="test very slow with tensorflow backend")
@pytest.mark.parametrize(
    "unbalanced_solver, divergence", itertools.product(["mm", "lbfgsb"], ["kl", "l2"])
)
def test_sanity(nx, unbalanced_solver, divergence):
    n_samples = 5  # nb samples

    mu_s = np.array([0, 0])
    cov_s = np.array([[1, 0], [0, 1]])

    xs = ot.datasets.make_2D_samples_gauss(n_samples, mu_s, cov_s, random_state=4)
    xt = xs[::-1].copy()

    px_s, px_f = ot.unif(n_samples), ot.unif(2)
    py_s, py_f = ot.unif(n_samples), ot.unif(2)

    xs_nx, xt_nx = nx.from_numpy(xs, xt)
    px_s_nx, px_f_nx, py_s_nx, py_f_nx = nx.from_numpy(px_s, px_f, py_s, py_f)

    reg_m = (10, 5)
    eps = 0
    max_iter_ot = 200
    max_iter = 200
    tol = 1e-7
    tol_ot = 1e-7

    # test couplings
    anti_id_sample = np.flipud(np.eye(n_samples, n_samples)) / n_samples
    id_feature = np.eye(2, 2) / 2

    pi_sample, pi_feature = unbalanced_co_optimal_transport(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=0,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    pi_sample_nx, pi_feature_nx = unbalanced_co_optimal_transport(
        X=xs_nx,
        Y=xt_nx,
        wx_samp=px_s_nx,
        wx_feat=px_f_nx,
        wy_samp=py_s_nx,
        wy_feat=py_f_nx,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=0,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )
    pi_sample_nx = nx.to_numpy(pi_sample_nx)
    pi_feature_nx = nx.to_numpy(pi_feature_nx)

    np.testing.assert_allclose(pi_sample, anti_id_sample, atol=1e-05)
    np.testing.assert_allclose(pi_sample_nx, pi_sample, atol=1e-06)
    np.testing.assert_allclose(pi_feature, pi_feature_nx, atol=1e-06)
    np.testing.assert_allclose(pi_feature, id_feature, atol=1e-05)

    # test divergence
    ucoot = unbalanced_co_optimal_transport2(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=0,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    ucoot_nx = unbalanced_co_optimal_transport2(
        X=xs_nx,
        Y=xt_nx,
        wx_samp=px_s_nx,
        wx_feat=px_f_nx,
        wy_samp=py_s_nx,
        wy_feat=py_f_nx,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=0,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    ucoot_nx = nx.to_numpy(ucoot_nx)
    np.testing.assert_allclose(ucoot, ucoot_nx, atol=1e-08)
    np.testing.assert_allclose(ucoot, 0, atol=1e-06)


@pytest.skip_backend("jax", reason="test very slow with jax backend")
@pytest.skip_backend("tf", reason="test very slow with tensorflow backend")
@pytest.mark.parametrize(
    "unbalanced_solver, divergence, eps",
    itertools.product(
        ["sinkhorn", "sinkhorn_log", "mm", "lbfgsb"], ["kl", "l2"], [0, 1]
    ),
)
def test_init_plans(nx, unbalanced_solver, divergence, eps):
    n_samples = 5  # nb samples

    mu_s = np.array([0, 0])
    cov_s = np.array([[1, 0], [0, 1]])

    xs = ot.datasets.make_2D_samples_gauss(n_samples, mu_s, cov_s, random_state=4)
    xt = xs[::-1].copy()

    px_s, px_f = ot.unif(n_samples), ot.unif(2)
    py_s, py_f = ot.unif(n_samples), ot.unif(2)
    G0_samp = px_s[:, None] * py_s[None, :]
    G0_feat = px_f[:, None] * py_f[None, :]

    xs_nx, xt_nx, G0_samp_nx, G0_feat_nx = nx.from_numpy(xs, xt, G0_samp, G0_feat)
    px_s_nx, px_f_nx, py_s_nx, py_f_nx = nx.from_numpy(px_s, px_f, py_s, py_f)

    reg_m = (1, 5)
    alpha = (0.1, 0.2)
    max_iter_ot = 5
    max_iter = 5
    tol = 1e-7
    tol_ot = 1e-7

    # test couplings
    pi_sample, pi_feature = unbalanced_co_optimal_transport(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    pi_sample_nx, pi_feature_nx = unbalanced_co_optimal_transport(
        X=xs_nx,
        Y=xt_nx,
        wx_samp=px_s_nx,
        wx_feat=px_f_nx,
        wy_samp=py_s_nx,
        wy_feat=py_f_nx,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=(G0_samp_nx, G0_feat_nx),
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )
    pi_sample_nx = nx.to_numpy(pi_sample_nx)
    pi_feature_nx = nx.to_numpy(pi_feature_nx)

    np.testing.assert_allclose(pi_sample_nx, pi_sample, atol=1e-03)
    np.testing.assert_allclose(pi_feature, pi_feature_nx, atol=1e-03)

    # test divergence
    ucoot = unbalanced_co_optimal_transport2(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    ucoot_nx = unbalanced_co_optimal_transport2(
        X=xs_nx,
        Y=xt_nx,
        wx_samp=px_s_nx,
        wx_feat=px_f_nx,
        wy_samp=py_s_nx,
        wy_feat=py_f_nx,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=(G0_samp_nx, G0_feat_nx),
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    ucoot_nx = nx.to_numpy(ucoot_nx)
    np.testing.assert_allclose(ucoot, ucoot_nx, atol=1e-08)


@pytest.skip_backend("jax", reason="test very slow with jax backend")
@pytest.skip_backend("tf", reason="test very slow with tensorflow backend")
@pytest.mark.parametrize(
    "unbalanced_solver, divergence, eps",
    itertools.product(
        ["sinkhorn", "sinkhorn_log", "mm", "lbfgsb"], ["kl", "l2"], [0, 1]
    ),
)
def test_init_duals(nx, unbalanced_solver, divergence, eps):
    n_samples = 5  # nb samples

    mu_s = np.array([0, 0])
    cov_s = np.array([[1, 0], [0, 1]])

    xs = ot.datasets.make_2D_samples_gauss(n_samples, mu_s, cov_s, random_state=4)
    xt = xs[::-1].copy()

    px_s, px_f = ot.unif(n_samples), ot.unif(2)
    py_s, py_f = ot.unif(n_samples), ot.unif(2)

    xs_nx, xt_nx = nx.from_numpy(xs, xt)
    px_s_nx, px_f_nx, py_s_nx, py_f_nx = nx.from_numpy(px_s, px_f, py_s, py_f)

    init_duals_samp = nx.from_numpy(np.zeros(n_samples), np.zeros(n_samples))
    init_duals_feat = nx.from_numpy(np.zeros(2), np.zeros(2))
    init_duals = (init_duals_samp, init_duals_feat)

    reg_m = (10, 5)
    alpha = (0.1, 0.2)
    max_iter_ot = 5
    max_iter = 5
    tol = 1e-7
    tol_ot = 1e-7

    # test couplings
    pi_sample, pi_feature = unbalanced_co_optimal_transport(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    pi_sample_nx, pi_feature_nx = unbalanced_co_optimal_transport(
        X=xs_nx,
        Y=xt_nx,
        wx_samp=px_s_nx,
        wx_feat=px_f_nx,
        wy_samp=py_s_nx,
        wy_feat=py_f_nx,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=init_duals,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )
    pi_sample_nx = nx.to_numpy(pi_sample_nx)
    pi_feature_nx = nx.to_numpy(pi_feature_nx)

    np.testing.assert_allclose(pi_sample_nx, pi_sample, atol=1e-03)
    np.testing.assert_allclose(pi_feature, pi_feature_nx, atol=1e-03)

    # test divergence
    ucoot = unbalanced_co_optimal_transport2(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    ucoot_nx = unbalanced_co_optimal_transport2(
        X=xs_nx,
        Y=xt_nx,
        wx_samp=px_s_nx,
        wx_feat=px_f_nx,
        wy_samp=py_s_nx,
        wy_feat=py_f_nx,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=init_duals,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    ucoot_nx = nx.to_numpy(ucoot_nx)
    np.testing.assert_allclose(ucoot, ucoot_nx, atol=1e-08)


@pytest.skip_backend("jax", reason="test very slow with jax backend")
@pytest.skip_backend("tf", reason="test very slow with tensorflow backend")
@pytest.mark.parametrize(
    "unbalanced_solver, divergence, eps",
    itertools.product(
        ["sinkhorn", "sinkhorn_log", "mm", "lbfgsb"], ["kl", "l2"], [0, 1e-2]
    ),
)
def test_linear_part(nx, unbalanced_solver, divergence, eps):
    n_samples = 5  # nb samples

    mu_s = np.array([0, 0])
    cov_s = np.array([[1, 0], [0, 1]])

    xs = ot.datasets.make_2D_samples_gauss(n_samples, mu_s, cov_s, random_state=4)
    xt = xs[::-1].copy()

    px_s, px_f = ot.unif(n_samples), ot.unif(2)
    py_s, py_f = ot.unif(n_samples), ot.unif(2)

    xs_nx, xt_nx = nx.from_numpy(xs, xt)
    px_s_nx, px_f_nx, py_s_nx, py_f_nx = nx.from_numpy(px_s, px_f, py_s, py_f)

    # linear part
    M_samp = np.ones((n_samples, n_samples))
    np.fill_diagonal(np.fliplr(M_samp), 0)
    M_feat = np.ones((2, 2))
    np.fill_diagonal(M_feat, 0)
    M_samp_nx, M_feat_nx = nx.from_numpy(M_samp, M_feat)

    reg_m = (10, 5)
    alpha = (0.1, 0.2)
    max_iter_ot = 5
    max_iter = 5
    tol = 1e-7
    tol_ot = 1e-7

    # test couplings
    pi_sample, pi_feature = unbalanced_co_optimal_transport(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=M_samp,
        M_feat=M_feat,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    pi_sample_nx, pi_feature_nx = unbalanced_co_optimal_transport(
        X=xs_nx,
        Y=xt_nx,
        wx_samp=px_s_nx,
        wx_feat=px_f_nx,
        wy_samp=py_s_nx,
        wy_feat=py_f_nx,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=M_samp_nx,
        M_feat=M_feat_nx,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )
    pi_sample_nx = nx.to_numpy(pi_sample_nx)
    pi_feature_nx = nx.to_numpy(pi_feature_nx)

    np.testing.assert_allclose(pi_sample_nx, pi_sample, atol=1e-06)
    np.testing.assert_allclose(pi_feature, pi_feature_nx, atol=1e-06)

    # test divergence
    ucoot = unbalanced_co_optimal_transport2(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=M_samp,
        M_feat=M_feat,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    ucoot_nx = unbalanced_co_optimal_transport2(
        X=xs_nx,
        Y=xt_nx,
        wx_samp=px_s_nx,
        wx_feat=px_f_nx,
        wy_samp=py_s_nx,
        wy_feat=py_f_nx,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=M_samp_nx,
        M_feat=M_feat_nx,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    ucoot_nx = nx.to_numpy(ucoot_nx)
    np.testing.assert_allclose(ucoot, ucoot_nx, atol=1e-08)


@pytest.skip_backend("jax", reason="test very slow with jax backend")
@pytest.skip_backend("tf", reason="test very slow with tensorflow backend")
@pytest.mark.parametrize(
    "unbalanced_solver, divergence, eps",
    itertools.product(
        ["sinkhorn", "sinkhorn_log", "mm", "lbfgsb"], ["kl", "l2"], [0, 1]
    ),
)
def test_reg_marginals(nx, unbalanced_solver, divergence, eps):
    n_samples = 5  # nb samples

    mu_s = np.array([0, 0])
    cov_s = np.array([[1, 0], [0, 1]])

    xs = ot.datasets.make_2D_samples_gauss(n_samples, mu_s, cov_s, random_state=4)
    xt = xs[::-1].copy()

    px_s, px_f = ot.unif(n_samples), ot.unif(2)
    py_s, py_f = ot.unif(n_samples), ot.unif(2)

    xs_nx, xt_nx = nx.from_numpy(xs, xt)
    px_s_nx, px_f_nx, py_s_nx, py_f_nx = nx.from_numpy(px_s, px_f, py_s, py_f)

    alpha = (0.1, 0.2)
    max_iter_ot = 5
    max_iter = 5
    tol = 1e-7
    tol_ot = 1e-7

    reg_m = 100
    full_list_reg_m = [reg_m, reg_m]
    full_tuple_reg_m = (reg_m, reg_m)
    tuple_reg_m, list_reg_m = (reg_m), [reg_m]

    list_options = [full_tuple_reg_m, tuple_reg_m, full_list_reg_m, list_reg_m]

    # test couplings
    pi_sample, pi_feature = unbalanced_co_optimal_transport(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    # test divergence
    ucoot = unbalanced_co_optimal_transport2(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    for opt in list_options:
        pi_sample_nx, pi_feature_nx = unbalanced_co_optimal_transport(
            X=xs_nx,
            Y=xt_nx,
            wx_samp=px_s_nx,
            wx_feat=px_f_nx,
            wy_samp=py_s_nx,
            wy_feat=py_f_nx,
            reg_marginals=opt,
            epsilon=eps,
            divergence=divergence,
            unbalanced_solver=unbalanced_solver,
            alpha=alpha,
            M_samp=None,
            M_feat=None,
            init_pi=None,
            init_duals=None,
            max_iter=max_iter,
            tol=tol,
            max_iter_ot=max_iter_ot,
            tol_ot=tol_ot,
            log=False,
            verbose=False,
        )
        pi_sample_nx = nx.to_numpy(pi_sample_nx)
        pi_feature_nx = nx.to_numpy(pi_feature_nx)

        np.testing.assert_allclose(pi_sample_nx, pi_sample, atol=1e-06)
        np.testing.assert_allclose(pi_feature, pi_feature_nx, atol=1e-06)

        # test divergence
        ucoot_nx = unbalanced_co_optimal_transport2(
            X=xs_nx,
            Y=xt_nx,
            wx_samp=px_s_nx,
            wx_feat=px_f_nx,
            wy_samp=py_s_nx,
            wy_feat=py_f_nx,
            reg_marginals=opt,
            epsilon=eps,
            divergence=divergence,
            unbalanced_solver=unbalanced_solver,
            alpha=alpha,
            M_samp=None,
            M_feat=None,
            init_pi=None,
            init_duals=None,
            max_iter=max_iter,
            tol=tol,
            max_iter_ot=max_iter_ot,
            tol_ot=tol_ot,
            method_sinkhorn="sinkhorn",
            log=False,
            verbose=False,
        )

        ucoot_nx = nx.to_numpy(ucoot_nx)
        np.testing.assert_allclose(ucoot, ucoot_nx, atol=1e-08)


@pytest.skip_backend("jax", reason="test very slow with jax backend")
@pytest.skip_backend("tf", reason="test very slow with tensorflow backend")
@pytest.mark.parametrize(
    "unbalanced_solver, divergence, alpha",
    itertools.product(
        ["sinkhorn", "sinkhorn_log", "mm", "lbfgsb"], ["kl", "l2"], [0, 1]
    ),
)
def test_eps(nx, unbalanced_solver, divergence, alpha):
    n_samples = 5  # nb samples

    mu_s = np.array([0, 0])
    cov_s = np.array([[1, 0], [0, 1]])

    xs = ot.datasets.make_2D_samples_gauss(n_samples, mu_s, cov_s, random_state=4)
    xt = xs[::-1].copy()

    px_s, px_f = ot.unif(n_samples), ot.unif(2)
    py_s, py_f = ot.unif(n_samples), ot.unif(2)

    xs_nx, xt_nx = nx.from_numpy(xs, xt)
    px_s_nx, px_f_nx, py_s_nx, py_f_nx = nx.from_numpy(px_s, px_f, py_s, py_f)

    reg_m = (10, 5)
    alpha = (0.1, 0.2)
    max_iter_ot = 5
    max_iter = 5
    tol = 1e-7
    tol_ot = 1e-7

    eps = 1
    full_list_eps = [eps, eps]
    full_tuple_eps = (eps, eps)
    tuple_eps, list_eps = (eps), [eps]

    list_options = [full_list_eps, full_tuple_eps, tuple_eps, list_eps]

    # test couplings
    pi_sample, pi_feature = unbalanced_co_optimal_transport(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    # test divergence
    ucoot = unbalanced_co_optimal_transport2(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    for opt in list_options:
        pi_sample_nx, pi_feature_nx = unbalanced_co_optimal_transport(
            X=xs_nx,
            Y=xt_nx,
            wx_samp=px_s_nx,
            wx_feat=px_f_nx,
            wy_samp=py_s_nx,
            wy_feat=py_f_nx,
            reg_marginals=reg_m,
            epsilon=opt,
            divergence=divergence,
            unbalanced_solver=unbalanced_solver,
            alpha=alpha,
            M_samp=None,
            M_feat=None,
            init_pi=None,
            init_duals=None,
            max_iter=max_iter,
            tol=tol,
            max_iter_ot=max_iter_ot,
            tol_ot=tol_ot,
            log=False,
            verbose=False,
        )
        pi_sample_nx = nx.to_numpy(pi_sample_nx)
        pi_feature_nx = nx.to_numpy(pi_feature_nx)

        np.testing.assert_allclose(pi_sample_nx, pi_sample, atol=1e-06)
        np.testing.assert_allclose(pi_feature, pi_feature_nx, atol=1e-06)

        # test divergence
        ucoot_nx = unbalanced_co_optimal_transport2(
            X=xs_nx,
            Y=xt_nx,
            wx_samp=px_s_nx,
            wx_feat=px_f_nx,
            wy_samp=py_s_nx,
            wy_feat=py_f_nx,
            reg_marginals=reg_m,
            epsilon=opt,
            divergence=divergence,
            unbalanced_solver=unbalanced_solver,
            alpha=alpha,
            M_samp=None,
            M_feat=None,
            init_pi=None,
            init_duals=None,
            max_iter=max_iter,
            tol=tol,
            max_iter_ot=max_iter_ot,
            tol_ot=tol_ot,
            log=False,
            verbose=False,
        )

        ucoot_nx = nx.to_numpy(ucoot_nx)
        np.testing.assert_allclose(ucoot, ucoot_nx, atol=1e-08)


@pytest.skip_backend("jax", reason="test very slow with jax backend")
@pytest.skip_backend("tf", reason="test very slow with tensorflow backend")
@pytest.mark.parametrize(
    "unbalanced_solver, divergence, eps",
    itertools.product(
        ["sinkhorn", "sinkhorn_log", "mm", "lbfgsb"], ["kl", "l2"], [0, 1e-2]
    ),
)
def test_alpha(nx, unbalanced_solver, divergence, eps):
    n_samples = 5  # nb samples

    mu_s = np.array([0, 0])
    cov_s = np.array([[1, 0], [0, 1]])

    xs = ot.datasets.make_2D_samples_gauss(n_samples, mu_s, cov_s, random_state=4)
    xt = xs[::-1].copy()

    px_s, px_f = ot.unif(n_samples), ot.unif(2)
    py_s, py_f = ot.unif(n_samples), ot.unif(2)

    xs_nx, xt_nx = nx.from_numpy(xs, xt)
    px_s_nx, px_f_nx, py_s_nx, py_f_nx = nx.from_numpy(px_s, px_f, py_s, py_f)

    # linear part
    M_samp = np.ones((n_samples, n_samples))
    np.fill_diagonal(np.fliplr(M_samp), 0)
    M_feat = np.ones((2, 2))
    np.fill_diagonal(M_feat, 0)
    M_samp_nx, M_feat_nx = nx.from_numpy(M_samp, M_feat)

    reg_m = (10, 5)
    max_iter_ot = 5
    max_iter = 5
    tol = 1e-7
    tol_ot = 1e-7

    alpha = 1
    full_list_alpha = [alpha, alpha]
    full_tuple_alpha = (alpha, alpha)
    tuple_alpha, list_alpha = (alpha), [alpha]

    list_options = [full_list_alpha, full_tuple_alpha, tuple_alpha, list_alpha]

    # test couplings
    pi_sample, pi_feature = unbalanced_co_optimal_transport(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=M_samp,
        M_feat=M_feat,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    # test divergence
    ucoot = unbalanced_co_optimal_transport2(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=M_samp,
        M_feat=M_feat,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    for opt in list_options:
        pi_sample_nx, pi_feature_nx = unbalanced_co_optimal_transport(
            X=xs_nx,
            Y=xt_nx,
            wx_samp=px_s_nx,
            wx_feat=px_f_nx,
            wy_samp=py_s_nx,
            wy_feat=py_f_nx,
            reg_marginals=reg_m,
            epsilon=eps,
            divergence=divergence,
            unbalanced_solver=unbalanced_solver,
            alpha=opt,
            M_samp=M_samp_nx,
            M_feat=M_feat_nx,
            init_pi=None,
            init_duals=None,
            max_iter=max_iter,
            tol=tol,
            max_iter_ot=max_iter_ot,
            tol_ot=tol_ot,
            log=False,
            verbose=False,
        )
        pi_sample_nx = nx.to_numpy(pi_sample_nx)
        pi_feature_nx = nx.to_numpy(pi_feature_nx)

        np.testing.assert_allclose(pi_sample_nx, pi_sample, atol=1e-06)
        np.testing.assert_allclose(pi_feature, pi_feature_nx, atol=1e-06)

        ucoot_nx = unbalanced_co_optimal_transport2(
            X=xs_nx,
            Y=xt_nx,
            wx_samp=px_s_nx,
            wx_feat=px_f_nx,
            wy_samp=py_s_nx,
            wy_feat=py_f_nx,
            reg_marginals=reg_m,
            epsilon=eps,
            divergence=divergence,
            unbalanced_solver=unbalanced_solver,
            alpha=opt,
            M_samp=M_samp_nx,
            M_feat=M_feat_nx,
            init_pi=None,
            init_duals=None,
            max_iter=max_iter,
            tol=tol,
            max_iter_ot=max_iter_ot,
            tol_ot=tol_ot,
            log=False,
            verbose=False,
        )

        ucoot_nx = nx.to_numpy(ucoot_nx)
        np.testing.assert_allclose(ucoot, ucoot_nx, atol=1e-08)


@pytest.skip_backend("jax", reason="test very slow with jax backend")
@pytest.skip_backend("tf", reason="test very slow with tensorflow backend")
@pytest.mark.parametrize(
    "unbalanced_solver, divergence, eps",
    itertools.product(
        ["sinkhorn", "sinkhorn_log", "mm", "lbfgsb"], ["kl", "l2"], [0, 1]
    ),
)
def test_log(nx, unbalanced_solver, divergence, eps):
    n_samples = 5  # nb samples

    mu_s = np.array([0, 0])
    cov_s = np.array([[1, 0], [0, 1]])

    xs = ot.datasets.make_2D_samples_gauss(n_samples, mu_s, cov_s, random_state=4)
    xt = xs[::-1].copy()

    px_s, px_f = ot.unif(n_samples), ot.unif(2)
    py_s, py_f = ot.unif(n_samples), ot.unif(2)

    xs_nx, xt_nx = nx.from_numpy(xs, xt)
    px_s_nx, px_f_nx, py_s_nx, py_f_nx = nx.from_numpy(px_s, px_f, py_s, py_f)

    reg_m = (10, 5)
    alpha = (0.1, 0.2)
    max_iter_ot = 5
    max_iter = 5
    tol = 1e-7
    tol_ot = 1e-7

    # test couplings
    pi_sample, pi_feature = unbalanced_co_optimal_transport(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    pi_sample_nx, pi_feature_nx, log = unbalanced_co_optimal_transport(
        X=xs_nx,
        Y=xt_nx,
        wx_samp=px_s_nx,
        wx_feat=px_f_nx,
        wy_samp=py_s_nx,
        wy_feat=py_f_nx,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=True,
        verbose=False,
    )
    pi_sample_nx = nx.to_numpy(pi_sample_nx)
    pi_feature_nx = nx.to_numpy(pi_feature_nx)

    np.testing.assert_allclose(pi_sample_nx, pi_sample, atol=1e-06)
    np.testing.assert_allclose(pi_feature, pi_feature_nx, atol=1e-06)

    # test divergence
    ucoot = unbalanced_co_optimal_transport2(
        X=xs,
        Y=xt,
        wx_samp=px_s,
        wx_feat=px_f,
        wy_samp=py_s,
        wy_feat=py_f,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    ucoot_nx = unbalanced_co_optimal_transport2(
        X=xs_nx,
        Y=xt_nx,
        wx_samp=px_s_nx,
        wx_feat=px_f_nx,
        wy_samp=py_s_nx,
        wy_feat=py_f_nx,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    ucoot_nx = nx.to_numpy(ucoot_nx)
    np.testing.assert_allclose(ucoot, ucoot_nx, atol=1e-08)


@pytest.skip_backend("jax", reason="test very slow with jax backend")
@pytest.skip_backend("tf", reason="test very slow with tensorflow backend")
@pytest.mark.parametrize(
    "unbalanced_solver, divergence, eps",
    itertools.product(
        ["sinkhorn", "sinkhorn_log", "mm", "lbfgsb"], ["kl", "l2"], [0, 1]
    ),
)
def test_marginals(nx, unbalanced_solver, divergence, eps):
    n_samples = 5  # nb samples

    mu_s = np.array([0, 0])
    cov_s = np.array([[1, 0], [0, 1]])

    xs = ot.datasets.make_2D_samples_gauss(n_samples, mu_s, cov_s, random_state=4)
    xt = xs[::-1].copy()

    px_s, px_f = ot.unif(n_samples), ot.unif(2)
    py_s, py_f = ot.unif(n_samples), ot.unif(2)

    xs_nx, xt_nx = nx.from_numpy(xs, xt)
    px_s_nx, px_f_nx, py_s_nx, py_f_nx = nx.from_numpy(px_s, px_f, py_s, py_f)

    reg_m = (10, 5)
    alpha = (0.1, 0.2)
    max_iter_ot = 5
    max_iter = 5
    tol = 1e-7
    tol_ot = 1e-7

    # test couplings
    pi_sample, pi_feature = unbalanced_co_optimal_transport(
        X=xs,
        Y=xt,
        wx_samp=None,
        wx_feat=None,
        wy_samp=None,
        wy_feat=None,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    pi_sample_nx, pi_feature_nx = unbalanced_co_optimal_transport(
        X=xs_nx,
        Y=xt_nx,
        wx_samp=px_s_nx,
        wx_feat=px_f_nx,
        wy_samp=py_s_nx,
        wy_feat=py_f_nx,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )
    pi_sample_nx = nx.to_numpy(pi_sample_nx)
    pi_feature_nx = nx.to_numpy(pi_feature_nx)

    np.testing.assert_allclose(pi_sample_nx, pi_sample, atol=1e-06)
    np.testing.assert_allclose(pi_feature, pi_feature_nx, atol=1e-06)

    # test divergence
    ucoot = unbalanced_co_optimal_transport2(
        X=xs,
        Y=xt,
        wx_samp=None,
        wx_feat=None,
        wy_samp=None,
        wy_feat=None,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    ucoot_nx = unbalanced_co_optimal_transport2(
        X=xs_nx,
        Y=xt_nx,
        wx_samp=px_s_nx,
        wx_feat=px_f_nx,
        wy_samp=py_s_nx,
        wy_feat=py_f_nx,
        reg_marginals=reg_m,
        epsilon=eps,
        divergence=divergence,
        unbalanced_solver=unbalanced_solver,
        alpha=alpha,
        M_samp=None,
        M_feat=None,
        init_pi=None,
        init_duals=None,
        max_iter=max_iter,
        tol=tol,
        max_iter_ot=max_iter_ot,
        tol_ot=tol_ot,
        log=False,
        verbose=False,
    )

    ucoot_nx = nx.to_numpy(ucoot_nx)
    np.testing.assert_allclose(ucoot, ucoot_nx, atol=1e-08)


@pytest.skip_backend("jax", reason="test very slow with jax backend")
@pytest.skip_backend("tf", reason="test very slow with tensorflow backend")
def test_raise_value_error(nx):
    n_samples = 5  # nb samples

    mu_s = np.array([0, 0])
    cov_s = np.array([[1, 0], [0, 1]])

    xs = ot.datasets.make_2D_samples_gauss(n_samples, mu_s, cov_s, random_state=4)
    xt = xs[::-1].copy()

    px_s, px_f = ot.unif(n_samples), ot.unif(2)
    py_s, py_f = ot.unif(n_samples), ot.unif(2)

    xs_nx, xt_nx = nx.from_numpy(xs, xt)
    px_s_nx, px_f_nx, py_s_nx, py_f_nx = nx.from_numpy(px_s, px_f, py_s, py_f)

    reg_m = (10, 5)
    eps = 0
    max_iter_ot = 5
    max_iter = 5
    tol = 1e-7
    tol_ot = 1e-7

    # raise error of divergence
    def ucoot_div(divergence):
        return unbalanced_co_optimal_transport(
            X=xs,
            Y=xt,
            wx_samp=px_s,
            wx_feat=px_f,
            wy_samp=py_s,
            wy_feat=py_f,
            reg_marginals=reg_m,
            epsilon=eps,
            divergence=divergence,
            unbalanced_solver="mm",
            alpha=0,
            M_samp=None,
            M_feat=None,
            init_pi=None,
            init_duals=None,
            max_iter=max_iter,
            tol=tol,
            max_iter_ot=max_iter_ot,
            tol_ot=tol_ot,
            log=False,
            verbose=False,
        )

    def ucoot_div_nx(divergence):
        return unbalanced_co_optimal_transport(
            X=xs_nx,
            Y=xt_nx,
            wx_samp=px_s_nx,
            wx_feat=px_f_nx,
            wy_samp=py_s_nx,
            wy_feat=py_f_nx,
            reg_marginals=reg_m,
            epsilon=eps,
            divergence=divergence,
            unbalanced_solver="mm",
            alpha=0,
            M_samp=None,
            M_feat=None,
            init_pi=None,
            init_duals=None,
            max_iter=max_iter,
            tol=tol,
            max_iter_ot=max_iter_ot,
            tol_ot=tol_ot,
            log=False,
            verbose=False,
        )

    np.testing.assert_raises(NotImplementedError, ucoot_div, "div_not_existed")
    np.testing.assert_raises(NotImplementedError, ucoot_div_nx, "div_not_existed")

    # raise error of solver
    def ucoot_solver(unbalanced_solver):
        return unbalanced_co_optimal_transport(
            X=xs,
            Y=xt,
            wx_samp=px_s,
            wx_feat=px_f,
            wy_samp=py_s,
            wy_feat=py_f,
            reg_marginals=reg_m,
            epsilon=eps,
            divergence="kl",
            unbalanced_solver=unbalanced_solver,
            alpha=0,
            M_samp=None,
            M_feat=None,
            init_pi=None,
            init_duals=None,
            max_iter=max_iter,
            tol=tol,
            max_iter_ot=max_iter_ot,
            tol_ot=tol_ot,
            log=False,
            verbose=False,
        )

    def ucoot_solver_nx(unbalanced_solver):
        return unbalanced_co_optimal_transport(
            X=xs_nx,
            Y=xt_nx,
            wx_samp=px_s_nx,
            wx_feat=px_f_nx,
            wy_samp=py_s_nx,
            wy_feat=py_f_nx,
            reg_marginals=reg_m,
            epsilon=eps,
            divergence="kl",
            unbalanced_solver=unbalanced_solver,
            alpha=0,
            M_samp=None,
            M_feat=None,
            init_pi=None,
            init_duals=None,
            max_iter=max_iter,
            tol=tol,
            max_iter_ot=max_iter_ot,
            tol_ot=tol_ot,
            log=False,
            verbose=False,
        )

    np.testing.assert_raises(NotImplementedError, ucoot_solver, "solver_not_existed")
    np.testing.assert_raises(NotImplementedError, ucoot_solver_nx, "solver_not_existed")



from ot.backend import get_backend


def entropy_batch(T, nx=None, eps=1e-16):
    r"""Computes the entropy of a batch of transport plans T.

    .. math::
        H(T)_b = - \sum_{i,j} T_{b,i,j} \log(T_{b,i,j})

    Parameters
    ----------

    T : array-like, shape (b,n1,n2)
        `b` transport of size `n1` x `n2`.
    eps : float, optional
        Small constant to avoid numerical issues with log(0). Default is 1e-16.
    nx : Backend, optional
        Backend to perform computations on. If omitted, the backend defaults to that of `T`.
    """
    if nx is None:
        nx = get_backend(T)
    return -nx.sum(T * nx.log(T + eps), axis=(1, 2))


def norm_batch(u, p=2, nx=None):
    """Computes the lp norm of a batch of vectors u."""
    if nx is None:
        nx = get_backend(u)
    return nx.sum(u**p, axis=1) ** (1 / p)


def bmv(A, b, nx):
    """
    Batched matrix-vector multiplication for tensor A and vector b.
    """
    return nx.einsum("bij,bj->bi", A, b)


def bop(a, b, nx):
    """
    Batched outer product for vectors a and b.
    """
    return nx.einsum("bi,bj->bij", a, b)


def bregman_projection_batch(
    K, a=None, b=None, nx=None, max_iter=10000, tol=1e-5, grad="detach"
):
    r"""
    Apply Bregman projection to a batch of affinity matrices :math:`\mathbf{K}`.

    The function solves the following optimization problem:

    .. math::
        \begin{aligned}
            \mathbf{T} = \mathop{\arg \min}_\mathbf{T} \quad & \text{KL}(\mathbf{T} \| \mathbf{K}) \\
            \text{s.t.} \quad & \mathbf{T} \mathbf{1} = \mathbf{a} \\
            & \mathbf{T}^T \mathbf{1} = \mathbf{b} \\
            & \mathbf{T} \geq 0
        \end{aligned}

    This is equivalent to:

    .. math::
        \begin{aligned}
            \mathbf{T} = \mathop{\arg \max}_\mathbf{T} \quad & \langle \mathbf{T},  \log(\mathbf{K}) \rangle_F \\
            \text{s.t.} \quad & \mathbf{T} \mathbf{1} = \mathbf{a} \\
            & \mathbf{T}^T \mathbf{1} = \mathbf{b} \\
            & \mathbf{T} \geq 0
        \end{aligned}

    The optimal solution has the form :math:`\mathbf{T} = \text{diag}(\mathbf{f}) \mathbf{K} \text{diag}(\mathbf{g})`,
    where the dual variables :math:`\mathbf{u}` and :math:`\mathbf{v}` are found iteratively using:

    .. math::
        \mathbf{f}^{(k+1)} = \frac{\mathbf{a}}{\sum \mathbf{K} \mathbf{g}^{(k)}}
        
        \mathbf{g}^{(k+1)} = \frac{\mathbf{b}}{\sum \mathbf{K}^T \mathbf{f}^{(k)}}

    Parameters
    ----------
    K : array-like, shape (B, n, m)
        Affinity matrix for each problem in the batch.
    a : array-like, shape (B, n), optional
        Source distribution for each problem. If None, uniform distribution is used.
    b : array-like, shape (B, m), optional
        Target distribution for each problem. If None, uniform distribution is used.
    nx : backend object, optional
        Numerical backend to use for computations. If None, the default backend is used.
    max_iter : int, optional
        Maximum number of iterations.
    tol : float, optional
        Tolerance for convergence. The solver stops when the maximum change in
        the dual variables is below this value.
    grad : str, optional
        Gradient computation mode: 'detach', 'autodiff', or 'last_step'.

    Returns
    -------
    dict
        Dictionary containing:
        - 'T' : array-like, shape (B, n, m)
            Optimal transport plan for each problem.
        - 'potentials' : tuple of array-like, shapes ((B, n), (B, m))
            Scaling factors (f, g).
        - 'n_iters' : int
            Number of iterations performed.


    Examples
    --------
    >>> import numpy as np
    >>> from ot.batch import bregman_projection_batch
    >>> # Create batch of affinity matrices
    >>> K = np.random.randn(5, 10, 15)  # 5 problems, 10x15 cost matrices
    >>> result = bregman_projection_batch(K)
    >>> T = result['T']  # Shape (5, 10, 15)

    See Also
    --------
    ot.batch.bregman_log_projection_batch : Bregman projection in the log-domain.
    """
    if nx is None:
        nx = get_backend(a, b, K)

    B, n, m = K.shape

    if a is None:
        a = nx.ones((B, n)) / n
    if b is None:
        b = nx.ones((B, m)) / m

    if grad == "detach":
        K = nx.detach(K)
    elif grad == "last_step":
        K_, K = K.clone(), nx.detach(K)

    f = nx.ones((B, n))  # a / nx.sum(K, axis=2)
    g = nx.ones((B, m))  # b / nx.sum(K, axis=1)

    for n_iters in range(max_iter):
        f = a / nx.sum(K * g[:, None, :], axis=2)
        g = b / nx.sum(K * f[:, :, None], axis=1)
        if n_iters % 10 == 0:
            T = K * f[:, :, None] * g[:, None, :]
            marginal = nx.sum(T, axis=2)
            err = nx.max(norm_batch(marginal - a))
            if err < tol:
                break

    if grad == "last_step":
        summand_f = nx.sum(K_ * g[:, None, :], axis=2)
        f = a / summand_f
        summand_g = nx.sum(K_ * f[:, :, None], axis=1)
        g = b / summand_g

    T = K * f[:, :, None] * g[:, None, :]
    potentials = (f, g)

    out = {"T": T, "n_iters": n_iters, "potentials": potentials}

    return out


def bregman_log_projection_batch(
    K, a=None, b=None, nx=None, max_iter=10000, tol=1e-5, grad="detach"
):
    r"""
    Apply Bregman projection to a batch of affinity matrices :math:`\exp(\mathbf{K})`.

    The function solves the following optimization problem:

    .. math::
        \begin{aligned}
            \mathbf{T} = \mathop{\arg \min}_\mathbf{T} \quad & \text{KL}(\mathbf{T} \| \exp(\mathbf{K})) \\
            \text{s.t.} \quad & \mathbf{T} \mathbf{1} = \mathbf{a} \\
            & \mathbf{T}^T \mathbf{1} = \mathbf{b} \\
            & \mathbf{T} \geq 0
        \end{aligned}

    This is equivalent to:

    .. math::
        \begin{aligned}
            \mathbf{T} = \mathop{\arg \max}_\mathbf{T} \quad & \langle \mathbf{T}, \mathbf{K} \rangle_F \\
            \text{s.t.} \quad & \mathbf{T} \mathbf{1} = \mathbf{a} \\
            & \mathbf{T}^T \mathbf{1} = \mathbf{b} \\
            & \mathbf{T} \geq 0
        \end{aligned}

    The optimal solution has the form :math:`\mathbf{T} = \text{diag}(\exp(\mathbf{u})) \exp(\mathbf{K}) \text{diag}(\exp(\mathbf{v}))`,
    where the dual variables :math:`\mathbf{u}` and :math:`\mathbf{v}` are found iteratively using:

    .. math::
        \mathbf{u}^{(k+1)} = \log(\mathbf{a}) - \text{LSE}(\mathbf{K} + \mathbf{v}^{(k)})

        \mathbf{v}^{(k+1)} = \log(\mathbf{b}) - \text{LSE}(\mathbf{K}^T + \mathbf{u}^{(k)})

    where LSE denotes the log-sum-exp operation. The iterations are performed using
    the log-sum-exp trick to avoid numerical issues.

    Parameters
    ----------
    K : array-like, shape (B, n, m)
        Affinity matrix for each problem in the batch.
    a : array-like, shape (B, n), optional
        Source distribution for each problem. If None, uniform distribution is used.
    b : array-like, shape (B, m), optional
        Target distribution for each problem. If None, uniform distribution is used.
    nx : backend object, optional
        Numerical backend to use for computations. If None, the default backend is used.
    max_iter : int, optional
        Maximum number of iterations.
    tol : float, optional
        Tolerance for convergence. The solver stops when the maximum change in
        the dual variables is below this value.
    grad : str, optional
        Gradient computation mode: 'detach', 'autodiff', or 'last_step'.

    Returns
    -------
    dict
        Dictionary containing:
        - 'T' : array-like, shape (B, n, m)
            Optimal transport plan for each problem.
        - 'log_T' : array-like, shape (B, n, m)
            Logarithm of the optimal transport plan for each problem.
        - 'potentials' : tuple of array-like, shapes ((B, n), (B, m))
            Log-scaling factors (u, v).
        - 'n_iters' : int
            Number of iterations performed.

    Examples
    --------
    >>> import numpy as np
    >>> from ot.batch import bregman_log_projection_batch
    >>> # Create batch of affinity matrices
    >>> K = np.random.randn(5, 10, 15)  # 5 problems, 10x15 cost matrices
    >>> result = bregman_log_projection_batch(K)
    >>> T = result['T']  # Shape (5, 10, 15)

    See Also
    --------
    ot.batch.bregman_projection_batch : standard Bregman projection.
    """

    if nx is None:
        nx = get_backend(a, b, K)

    B, n, m = K.shape

    if a is None:
        a = nx.ones((B, n)) / n
    if b is None:
        b = nx.ones((B, m)) / m

    u = nx.zeros((B, n), type_as=K)  # u = nx.log(a) - nx.logsumexp(K, axis=2).squeeze()
    v = nx.zeros((B, m), type_as=K)  # v = nx.log(b) - nx.logsumexp(K, axis=1).squeeze()

    loga = nx.log(a)
    logb = nx.log(b)

    if grad == "detach":
        K = nx.detach(K)
    elif grad == "last_step":
        K_, K = K.clone(), nx.detach(K)

    for n_iters in range(max_iter):
        u = loga - nx.logsumexp(K + v[:, None, :], axis=2)
        v = logb - nx.logsumexp(K + u[:, :, None], axis=1)

        # Check convergence once every 10 iterations
        if n_iters % 10 == 0:
            T = nx.exp(K + u[:, :, None] + v[:, None, :])
            marginal = nx.sum(T, axis=2)
            err = nx.max(norm_batch(marginal - a))
            if err < tol:
                break

    if grad == "last_step":
        summand_u = K_ + v[:, None, :]
        u = loga - nx.logsumexp(summand_u, axis=2)
        summand_v = K_ + u[:, :, None]
        v = logb - nx.logsumexp(summand_v, axis=1)

    log_T = K + u[:, :, None] + v[:, None, :]
    T = nx.exp(log_T)
    log_potentials = (u, v)

    return {"T": T, "log_T": log_T, "n_iters": n_iters, "potentials": log_potentials}



# -*- coding: utf-8 -*-
"""
Optimal transport for Gaussian Mixtures
"""

# Author: Eloi Tanguy <eloi.tanguy@math.cnrs.fr>
#         Remi Flamary <remi.flamary@polytechnique.edu>
#         Julie Delon <julie.delon@math.cnrs.fr>
#
# License: MIT License

from .backend import get_backend
from .lp import emd2, emd
import numpy as np
from .utils import dist
from .gaussian import bures_wasserstein_mapping, bures_wasserstein_barycenter


def gaussian_logpdf(x, m, C):
    r"""
    Compute the log of the probability density function of a multivariate
    Gaussian distribution.

    Parameters
    ----------
    x : array-like, shape (..., d)
        The input samples.
    m : array-like, shape (d,)
        The mean vector of the Gaussian distribution.
    C : array-like, shape (d, d)
        The covariance matrix of the Gaussian distribution.

    Returns
    -------
    pdf : array-like, shape (...,)
        The probability density function evaluated at each sample.

    """
    assert (
        x.shape[-1] == m.shape[-1] == C.shape[-1] == C.shape[-2]
    ), "Dimension mismatch"
    nx = get_backend(x, m, C)
    d = m.shape[0]
    diff = x - m
    inv_C = nx.inv(C)
    z = nx.sum(diff * (diff @ inv_C), axis=-1)
    _, log_det_C = nx.slogdet(C)
    return -0.5 * (d * np.log(2 * np.pi) + log_det_C + z)


def gaussian_pdf(x, m, C):
    r"""
    Compute the probability density function of a multivariate
    Gaussian distribution.

    Parameters
    ----------
    x : array-like, shape (..., d)
        The input samples.
    m : array-like, shape (d,)
        The mean vector of the Gaussian distribution.
    C : array-like, shape (d, d)
        The covariance matrix of the Gaussian distribution.

    Returns
    -------
    pdf : array-like, shape (...,)
        The probability density function evaluated at each sample.

    """
    return get_backend(x, m, C).exp(gaussian_logpdf(x, m, C))


def gmm_pdf(x, m, C, w):
    r"""
    Compute the probability density function (PDF) of a
    Gaussian Mixture Model (GMM) at given points.

    Parameters
    ----------
    x : array-like, shape (..., d)
        The input samples.
    m : array-like, shape (n_components, d)
        The means of the Gaussian components.
    C : array-like, shape (n_components, d, d)
        The covariance matrices of the Gaussian components.
    w : array-like, shape (n_components,)
        The weights of the Gaussian components.

    Returns
    -------
    out : array-like, shape (...,)
        The PDF values at the given points.

    """
    assert (
        m.shape[0] == C.shape[0] == w.shape[0]
    ), "All GMM parameters must have the same amount of components"
    nx = get_backend(x, m, C, w)
    out = nx.zeros((x.shape[:-1]))
    for k in range(m.shape[0]):
        out = out + w[k] * gaussian_pdf(x, m[k], C[k])
    return out


def dist_bures_squared(m_s, m_t, C_s, C_t):
    r"""
    Compute the matrix of the squared Bures distances between the components of
    two Gaussian Mixture Models (GMMs). Used to compute the GMM Optimal
    Transport distance [69].

    Parameters
    ----------
    m_s : array-like, shape (k_s, d)
        Mean vectors of the source GMM.
    m_t : array-like, shape (k_t, d)
        Mean vectors of the target GMM.
    C_s : array-like, shape (k_s, d, d)
        Covariance matrices of the source GMM.
    C_t : array-like, shape (k_t, d, d)
        Covariance matrices of the target GMM.

    Returns
    -------
    dist : array-like, shape (k_s, k_t)
        Matrix of squared Bures distances between the components of the source
        and target GMMs.

    References
    ----------
    .. [69] Delon, J., & Desolneux, A. (2020). A Wasserstein-type distance in the space of Gaussian mixture models. SIAM Journal on Imaging Sciences, 13(2), 936-970.

    """
    nx = get_backend(m_s, C_s, m_t, C_t)

    assert m_s.shape[0] == C_s.shape[0], "Source GMM has different amount of components"

    assert m_t.shape[0] == C_t.shape[0], "Target GMM has different amount of components"

    assert (
        m_s.shape[-1] == m_t.shape[-1] == C_s.shape[-1] == C_t.shape[-1]
    ), "All GMMs must have the same dimension"

    D_means = dist(m_s, m_t, metric="sqeuclidean")

    # C2[i, j] = Cs12[i] @ C_t[j] @ Cs12[i], shape (k_s, k_t, d, d)
    Cs12 = nx.sqrtm(C_s)  # broadcasts matrix sqrt over (k_s,)
    C2 = nx.einsum("ikl,jlm,imn->ijkn", Cs12, C_t, Cs12)
    C = nx.sqrtm(C2)  # broadcasts matrix sqrt over (k_s, k_t)

    # D_covs[i,j] = trace(C_s[i] + C_t[j] - 2C[i,j])
    trace_C_s = nx.einsum("ikk->i", C_s)[:, None]  # (k_s, 1)
    trace_C_t = nx.einsum("ikk->i", C_t)[None, :]  # (1, k_t)
    D_covs = trace_C_s + trace_C_t  # broadcasts to (k_s, k_t)
    D_covs -= 2 * nx.einsum("ijkk->ij", C)

    return nx.maximum(D_means + D_covs, 0)


def gmm_ot_loss(m_s, m_t, C_s, C_t, w_s, w_t, log=False):
    r"""
    Compute the Gaussian Mixture Model (GMM) Optimal Transport distance between
    two GMMs introduced in [69].

    Parameters
    ----------
    m_s : array-like, shape (k_s, d)
        Mean vectors of the source GMM.
    m_t : array-like, shape (k_t, d)
        Mean vectors of the target GMM.
    C_s : array-like, shape (k_s, d, d)
        Covariance matrices of the source GMM.
    C_t : array-like, shape (k_t, d, d)
        Covariance matrices of the target GMM.
    w_s : array-like, shape (k_s,)
        Weights of the source GMM components.
    w_t : array-like, shape (k_t,)
        Weights of the target GMM components.
    log: bool, optional (default=False)
        If True, returns a dictionary containing the cost and dual variables.
        Otherwise returns only the GMM optimal transportation cost.

    Returns
    -------
    loss : float or array-like
        The GMM-OT loss.
    log : dict, optional
        If input log is true, a dictionary containing the
        cost and dual variables and exit status

    References
    ----------
    .. [69] Delon, J., & Desolneux, A. (2020). A Wasserstein-type distance in the space of Gaussian mixture models. SIAM Journal on Imaging Sciences, 13(2), 936-970.

    """
    get_backend(m_s, C_s, w_s, m_t, C_t, w_t)

    assert m_s.shape[0] == w_s.shape[0], "Source GMM has different amount of components"

    assert m_t.shape[0] == w_t.shape[0], "Target GMM has different amount of components"

    D = dist_bures_squared(m_s, m_t, C_s, C_t)
    return emd2(w_s, w_t, D, log=log)


def gmm_ot_plan(m_s, m_t, C_s, C_t, w_s, w_t, log=False):
    r"""
    Compute the Gaussian Mixture Model (GMM) Optimal Transport plan between
    two GMMs introduced in [69].

    Parameters
    ----------
    m_s : array-like, shape (k_s, d)
        Mean vectors of the source GMM.
    m_t : array-like, shape (k_t, d)
        Mean vectors of the target GMM.
    C_s : array-like, shape (k_s, d, d)
        Covariance matrices of the source GMM.
    C_t : array-like, shape (k_t, d, d)
        Covariance matrices of the target GMM.
    w_s : array-like, shape (k_s,)
        Weights of the source GMM components.
    w_t : array-like, shape (k_t,)
        Weights of the target GMM components.
    log : bool, optional (default=False)
        If True, returns a dictionary containing the cost and dual variables.
        Otherwise returns only the GMM optimal transportation matrix.

    Returns
    -------
    plan : array-like, shape (k_s, k_t)
        The GMM-OT plan.
    log : dict, optional
        If input log is true, a dictionary containing the
        cost and dual variables and exit status

    References
    ----------
    .. [69] Delon, J., & Desolneux, A. (2020). A Wasserstein-type distance in the space of Gaussian mixture models. SIAM Journal on Imaging Sciences, 13(2), 936-970.

    """
    get_backend(m_s, C_s, w_s, m_t, C_t, w_t)

    assert m_s.shape[0] == w_s.shape[0], "Source GMM has different amount of components"

    assert m_t.shape[0] == w_t.shape[0], "Target GMM has different amount of components"

    D = dist_bures_squared(m_s, m_t, C_s, C_t)
    return emd(w_s, w_t, D, log=log)


def gmm_ot_apply_map(
    x, m_s, m_t, C_s, C_t, w_s, w_t, plan=None, method="bary", seed=None
):
    r"""
    Apply Gaussian Mixture Model (GMM) optimal transport (OT) mapping to input
    data. The 'barycentric' mapping corresponds to the barycentric projection
    of the GMM-OT plan, and is called T_bary in [69]. The 'random' mapping takes
    for each input point a random pair (i,j) of components of the GMMs and
    applied the Gaussian map, it is called T_rand in [69].

    Parameters
    ----------
    x : array-like, shape (n_samples, d)
        Input data points.
    m_s : array-like, shape (k_s, d)
        Mean vectors of the source GMM components.
    m_t : array-like, shape (k_t, d)
        Mean vectors of the target GMM components.
    C_s : array-like, shape (k_s, d, d)
        Covariance matrices of the source GMM components.
    C_t : array-like, shape (k_t, d, d)
        Covariance matrices of the target GMM components.
    w_s : array-like, shape (k_s,)
        Weights of the source GMM components.
    w_t : array-like, shape (k_t,)
        Weights of the target GMM components.
    plan : array-like, shape (k_s, k_t), optional
        Optimal transport plan between the source and target GMM components.
        If not provided, it will be computed internally.
    method : {'bary', 'rand'}, optional
        Method for applying the GMM OT mapping. 'bary' uses barycentric mapping,
        while 'rand' uses random sampling. Default is 'bary'.
    seed : int, optional
        Seed for the random number generator. Only used when method='rand'.

    Returns
    -------
    out : array-like, shape (n_samples, d)
        Output data points after applying the GMM OT mapping.

    References
    ----------
    .. [69] Delon, J., & Desolneux, A. (2020). A Wasserstein-type distance in the space of Gaussian mixture models. SIAM Journal on Imaging Sciences, 13(2), 936-970.

    """

    if plan is None:
        plan = gmm_ot_plan(m_s, m_t, C_s, C_t, w_s, w_t)
        nx = get_backend(x, m_s, m_t, C_s, C_t, w_s, w_t)
    else:
        nx = get_backend(x, m_s, m_t, C_s, C_t, w_s, w_t, plan)

    k_s, k_t = m_s.shape[0], m_t.shape[0]
    d = m_s.shape[1]
    n_samples = x.shape[0]

    if method == "bary":
        out = nx.zeros(x.shape)
        logpdf = nx.stack(
            [gaussian_logpdf(x, m_s[k], C_s[k])[:, None] for k in range(k_s)]
        )

        # only need to compute for non-zero plan entries
        for i, j in zip(*nx.where(plan > 0)):
            Cs12 = nx.sqrtm(C_s[i])
            Cs12inv = nx.inv(Cs12)

            M0 = nx.sqrtm(Cs12 @ C_t[j] @ Cs12)
            A = Cs12inv @ M0 @ Cs12inv
            b = m_t[j] - A @ m_s[i]

            # gaussian mapping between components i and j applied to x
            T_ij_x = x @ A + b
            z = w_s[:, None, None] * nx.exp(logpdf - logpdf[i][None, :, :])
            denom = nx.sum(z, axis=0)

            out = out + plan[i, j] * T_ij_x / denom

        return out

    else:  # rand
        # A[i, j] is the linear part of the gaussian mapping between components
        # i and j, b[i, j] is the translation part
        rng = np.random.RandomState(seed)

        A = nx.zeros((k_s, k_t, d, d))
        b = nx.zeros((k_s, k_t, d))

        # only need to compute for non-zero plan entries
        for i, j in zip(*nx.where(plan > 0)):
            Cs12 = nx.sqrtm(C_s[i])
            Cs12inv = nx.inv(Cs12)

            M0 = nx.sqrtm(Cs12 @ C_t[j] @ Cs12)
            A[i, j] = Cs12inv @ M0 @ Cs12inv
            b[i, j] = m_t[j] - A[i, j] @ m_s[i]

        logpdf = nx.stack(
            [gaussian_logpdf(x, m_s[k], C_s[k]) for k in range(k_s)], axis=-1
        )
        # (n_samples, k_s)
        out = nx.zeros(x.shape)

        for i_sample in range(n_samples):
            log_g = logpdf[i_sample]
            log_diff = log_g[:, None] - log_g[None, :]
            weighted_exp = w_s[:, None] * nx.exp(log_diff)
            denom = nx.sum(weighted_exp, axis=0)[:, None] * nx.ones(plan.shape[1])
            p_mat = plan / denom

            p = p_mat.reshape(k_s * k_t)  # stack line-by-line
            # sample between 0 and k_s * k_t - 1
            ij_mat = rng.choice(k_s * k_t, p=p)
            i = ij_mat // k_t
            j = ij_mat % k_t
            out[i_sample] = A[i, j] @ x[i_sample] + b[i, j]

        return out


def gmm_ot_plan_density(x, y, m_s, m_t, C_s, C_t, w_s, w_t, plan=None, atol=1e-2):
    """
    Compute the density of the Gaussian Mixture Model - Optimal Transport
    coupling between GMMS at given points, as introduced in [69].
    Given two arrays of points x and y, the function computes the density at
    each point `(x[i], y[i])` of the product space.

    Parameters
    ----------
    x : array-like, shape (n, d)
        Entry points in source space for density computation.
    y : array-like, shape (m, d)
        Entry points in target space for density computation.
    m_s : array-like, shape (k_s, d)
        The means of the source GMM components.
    m_t : array-like, shape (k_t, d)
        The means of the target GMM components.
    C_s : array-like, shape (k_s, d, d)
        The covariance matrices of the source GMM components.
    C_t : array-like, shape (k_t, d, d)
        The covariance matrices of the target GMM components.
    w_s : array-like, shape (k_s,)
        The weights of the source GMM components.
    w_t : array-like, shape (k_t,)
        The weights of the target GMM components.
    plan : array-like, shape (k_s, k_t), optional
        The optimal transport plan between the source and target GMMs.
        If not provided, it will be computed using `gmm_ot_plan`.
    atol : float, optional
        The absolute tolerance used to determine the support of the GMM-OT
        coupling.

    Returns
    -------
    density : array-like, shape (n, m)
        The density of the GMM-OT coupling between the two GMMs.

    References
    ----------
    .. [69] Delon, J., & Desolneux, A. (2020). A Wasserstein-type distance in the space of Gaussian mixture models. SIAM Journal on Imaging Sciences, 13(2), 936-970.

    """
    assert (
        x.shape[-1] == y.shape[-1]
    ), "x (n, d) and y (m, d) must have the same dimension d"
    n, m = x.shape[0], y.shape[0]
    nx = get_backend(x, y, m_s, m_t, C_s, C_t, w_s, w_t)

    # hand-made d-variate meshgrid in ij indexing
    xx = x[:, None, :] * nx.ones((1, m, 1))  # shapes (n, m, d)
    yy = y[None, :, :] * nx.ones((n, 1, 1))  # shapes (n, m, d)

    if plan is None:
        plan = gmm_ot_plan(m_s, m_t, C_s, C_t, w_s, w_t)

    def Tk0k1(k0, k1):
        A, b = bures_wasserstein_mapping(m_s[k0], m_t[k1], C_s[k0], C_t[k1])
        Tx = xx @ A + b
        g = gaussian_pdf(xx, m_s[k0], C_s[k0])
        out = plan[k0, k1] * g
        norms = nx.norm(Tx - yy, axis=-1)
        out = out * ((norms < atol) * 1.0)
        return out

    mat = nx.stack(
        [
            nx.stack([Tk0k1(k0, k1) for k1 in range(m_t.shape[0])])
            for k0 in range(m_s.shape[0])
        ]
    )
    return nx.sum(mat, axis=(0, 1))


def gmm_barycenter_fixed_point(
    means_list,
    covs_list,
    w_list,
    means_init,
    covs_init,
    weights,
    w_bar=None,
    iterations=100,
    log=False,
    barycentric_proj_method="euclidean",
):
    r"""
    Solves the Gaussian Mixture Model OT barycenter problem (defined in [69])
    using the fixed point algorithm (proposed in [77]). The
    weights of the barycenter are not optimized, and stay the same as the input
    `w_list` or are initialized to uniform.

    The algorithm uses barycentric projections of GMM-OT plans, and these can be
    computed either through Bures Barycenters (slow but accurate,
    barycentric_proj_method='bures') or by convex combination (fast,
    barycentric_proj_method='euclidean', default).

    This is a special case of the generic free-support barycenter solver
    `ot.lp.free_support_barycenter_generic_costs`.

    Parameters
    ----------
    means_list : list of array-like
        List of K (m_k, d) GMM means.
    covs_list : list of array-like
        List of K (m_k, d, d) GMM covariances.
    w_list : list of array-like
        List of K (m_k) arrays of weights.
    means_init : array-like
        Initial (n, d) GMM means.
    covs_init : array-like
        Initial (n, d, d) GMM covariances.
    weights : array-like
        Array (K,) of the barycentre coefficients.
    w_bar : array-like, optional
        Initial weights (n) of the barycentre GMM. If None, initialized to uniform.
    iterations : int, optional
        Number of iterations (default is 100).
    log : bool, optional
        Whether to return the list of iterations (default is False).
    barycentric_proj_method : str, optional
        Method to project the barycentre weights: 'euclidean' (default) or 'bures'.

    Returns
    -------
    means : array-like
        (n, d) barycentre GMM means.
    covs : array-like
        (n, d, d) barycentre GMM covariances.
    log_dict : dict, optional
        Dictionary containing the list of iterations if log is True.

    References
    ----------
    .. [69] Delon, J., & Desolneux, A. (2020). A Wasserstein-type distance in the space of Gaussian mixture models. SIAM Journal on Imaging Sciences, 13(2), 936-970.

    .. [77] Tanguy, Eloi and Delon, Julie and Gozlan, Nathaël (2024). Computing barycenters of Measures for Generic Transport Costs. arXiv preprint 2501.04016 (2024)

    See Also
    --------
    ot.lp.free_support_barycenter_generic_costs : Compute barycenter of measures for generic transport costs.
    """
    nx = get_backend(
        means_init, covs_init, means_list[0], covs_list[0], w_list[0], weights
    )
    K = len(means_list)
    n = means_init.shape[0]
    d = means_init.shape[1]
    means_its = [nx.copy(means_init)]
    covs_its = [nx.copy(covs_init)]
    means, covs = means_init, covs_init

    if w_bar is None:
        w_bar = nx.ones(n, type_as=means) / n

    for _ in range(iterations):
        pi_list = [
            gmm_ot_plan(means, means_list[k], covs, covs_list[k], w_bar, w_list[k])
            for k in range(K)
        ]

        # filled in the euclidean case
        means_selection, covs_selection = None, None

        # in the euclidean case, the selection of Gaussians from each K sources
        # comes from a barycentric projection: it is a convex combination of the
        # selected means and covariances, which can be computed without a
        # for loop on i = 0, ..., n -1
        if barycentric_proj_method == "euclidean":
            means_selection = nx.zeros((n, K, d), type_as=means)
            covs_selection = nx.zeros((n, K, d, d), type_as=means)
            for k in range(K):
                means_selection[:, k, :] = n * pi_list[k] @ means_list[k]
                covs_selection[:, k, :, :] = (
                    nx.einsum("ij,jab->iab", pi_list[k], covs_list[k]) * n
                )

        # each component i of the barycentre will be a Bures barycentre of the
        # selected components of the K GMMs. In the 'bures' barycentric
        # projection option, the selected components are also Bures barycentres.
        for i in range(n):
            # means_selection_i (K, d) is the selected means, each comes from a
            # Gaussian barycentre along the disintegration of pi_k at i
            # covs_selection_i (K, d, d) are the selected covariances
            means_selection_i = None
            covs_selection_i = None

            # use previous computation (convex combination)
            if barycentric_proj_method == "euclidean":
                means_selection_i = means_selection[i]
                covs_selection_i = covs_selection[i]

            # compute Bures barycentre of certain components to get the
            # selection at i
            elif barycentric_proj_method == "bures":
                means_selection_i = nx.zeros((K, d), type_as=means)
                covs_selection_i = nx.zeros((K, d, d), type_as=means)
                for k in range(K):
                    w = (1 / w_bar[i]) * pi_list[k][i, :]
                    m, C = bures_wasserstein_barycenter(means_list[k], covs_list[k], w)
                    means_selection_i[k] = m
                    covs_selection_i[k] = C

            else:
                raise ValueError("Unknown barycentric_proj_method")

            means[i], covs[i] = bures_wasserstein_barycenter(
                means_selection_i, covs_selection_i, weights
            )

        if log:
            means_its.append(nx.copy(means))
            covs_its.append(nx.copy(covs))

    if log:
        return means, covs, {"means_its": means_its, "covs_its": covs_its}
    return means, covs
